"""
.. currentmodule:: skrf.constants

========================================
constants (:mod:`skrf.constants`)
========================================

This module contains constants, numerical approximations, and unit conversions

.. data:: c

    Velocity of light constant (from scipy)

.. data:: INF

    A very very large value (1e99)

.. data:: ONE

    1 + epsilon where epsilon is small. Used to avoid numerical error.

.. data:: ZERO

    0 + epsilon where epsilon is small. Used to avoid numerical error.

.. data:: K_BOLTZMANN

    Boltzmann constant (1.38064852e-23)

.. data:: S_DEFINITIONS

    S-parameter definition labels:
        - 'power' for power-waves definition,
        - 'pseudo' for pseudo-waves definition.
        - 'traveling' corresponds to the initial implementation.

.. data:: S_DEF_DEFAULT

    Default S-parameter definition: 'power', for power-wave definition.

.. data:: S_DEF_HFSS_DEFAULT

    Default S-parameter definition setting in HFSS: 'traveling' ("Generalized" in HFSS)

.. autosummary::
   :toctree: generated/

   to_meters

"""
from __future__ import annotations

from numbers import Number
from typing import Literal, Sequence, Union, get_args

import numpy as np
from scipy.constants import c, inch, mil

# used as substitutes to handle mathematical singularities.
INF = 1e99
"""
High but not infinite value for numerical purposes.
"""

ALMOST_ZERO = 1e-12
"""
Very tiny but not zero value to handle mathematical singularities.
"""

ZERO = 1e-4
"""
A very small values, often used for numerical comparisons.
"""

ONE = 1.0 + 1/1e14
"""
Almost one but not one to handle mathematical singularities.
"""

LOG_OF_NEG = -100
"""
Very low but minus infinity value for numerical purposes.
"""

K_BOLTZMANN = 1.38064852e-23
"""
Boltzmann constant (1.38064852e-23)
"""

T0 = 290.
"""
Room temperature (kind of)
"""

EIG_COND = 1e-9
"""
Eigenvalue ratio compared to the maximum eigenvalue in :meth:`~skrf.mathFunctions.nudge_eig`.

EIG_COND * max(eigenvalue)
"""

EIG_MIN = 1e-12
"""
Minimum eigenvalue used in :meth:`~skrf.mathFunctions.nudge_eig`
"""

# S-parameter definition labels and default definition
SdefT = Literal["power", "pseudo", "traveling"]
S_DEFINITIONS: list[SdefT] = list(get_args(SdefT))
S_DEF_DEFAULT = 'power'
S_DEF_HFSS_DEFAULT = 'traveling'

FrequencyUnitT = Literal["Hz", "kHz", "MHz", "GHz", "THz"]
FREQ_UNITS: dict[FrequencyUnitT, float] = {"Hz": 1.0, "kHz": 1e3, "MHz": 1e6, "GHz": 1e9, "THz": 1e12}

SweepTypeT = Literal["lin", "log"]
CoordT = Literal["cart", "polar"]
InterpolKindT = Literal["linear", "cubic", "nearest", "zero", "slinear", "quadratic", "rational"]
PrimaryPropertiesT = Literal['s', 'z', 'y', 'a', 'h', 't']
ComponentFuncT = Literal["re", "im", "mag", "db", "db10", "rad", "deg", "arcl", "rad_unwrap", "deg_unwrap",
                         "arcl_unwrap", "vswr", "time", "time_db", "time_mag", "time_impulse", "time_step"]
SparamFormatT = Literal["db", "ri", "ma"]
PortOrderT = Literal["first", "second", "third", "last", "auto"]
CircuitComponentT = Literal["_is_circuit_port", "_is_circuit_ground", "_is_circuit_open"]
MemoryLayoutT = Literal["C", "F"]

NumberLike = Union[Number, Sequence[Number], np.ndarray]

global distance_dict
distance_dict = {
    'm': 1.,
    'cm': 1e-2,
    'mm': 1e-3,
    'um': 1e-6,
    'in': inch,
    'mil': mil,
    's': c,
    'us': 1e-6*c,
    'ns': 1e-9*c,
    'ps': 1e-12*c,
}


def to_meters(d: NumberLike, unit: str = 'm', v_g: float = c) -> NumberLike:
    """
    Translate various units of distance into meters.

    Parameters
    ----------
    d : number or array-like
        value(s) to convert
    unit : str
        the unit to that x is in:
        ['m','cm','um','in','mil','s','us','ns','ps']
    v_g : float
        group velocity in m/s

    Returns
    -------
    d_m : number of array-like
        distance in meters
    """
    _distance_dict = {
        'm': 1.,
        'cm': 1e-2,
        'mm': 1e-3,
        'um': 1e-6,
        'in': inch,
        'mil': mil,
        's': v_g,
        'us': 1e-6*v_g,
        'ns': 1e-9*v_g,
        'ps': 1e-12*v_g,
    }

    unit = unit.lower()
    try:
        return _distance_dict[unit]*d
    except KeyError as err:
        raise(ValueError('Incorrect unit')) from err



#
# scikit-rf (aka skrf) documentation build configuration file
#

# -- Project information -----------------------------------------------------

project = 'scikit-rf'
copyright = '2024, scikit-rf team'
author = 'scikit-rf team'

# -- General configuration ---------------------------------------------------

import os
import sys
import warnings

warnings.filterwarnings('ignore')

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('../../'))

import skrf as rf

rf.setup_pylab()

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
sys.path.insert(0, os.path.abspath('../sphinxext'))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'sphinx_rtd_theme',
    'nbsphinx',
    #'inheritance_diagram',
    'IPython.sphinxext.ipython_directive',
    'IPython.sphinxext.ipython_console_highlighting',
    ]

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True

# NBsphinx settings
nbsphinx_execute = 'always'
nbsphinx_allow_errors = True
nbsphinx_kernel_name = 'python'
numpydoc_show_class_members = False
nbsphinx_timeout = 180

# Autodoc settings
autodoc_typehints = "description"

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', "**/*.rst.rst", '**.ipynb_checkpoints']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.

html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = 'scikit-rf Documentation'

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = html_title

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = 'scikit-rf-logo-flat.svg'
#html_logo = '_static/scikit-rf-logo-flat-docs.svg'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = 'skrfdoc'


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
#latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
#latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ('index', 'scikit-rf.tex', 'scikit-rf Documentation',
   'scikit-rf Developers', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = '_static/scikit-rf-title-flat.pdf'

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Additional stuff for the LaTeX preamble.
latex_preamble = r'\usepackage{epstopdf}'

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True

autosummary_generate = True

# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'scikit-rf', 'scikit-rf Documentation',
     ['alex arsenovic'], 1)
]


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('https://numpy.org/doc/stable', None),
    'pd': ('https://pandas.pydata.org/docs', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/', None),
}

def process_signature(app, what, name, obj, options, signature, return_annotation):

    func_name = name.split(".")[-1]
    if func_name.startswith("plot_") and func_name[5:] in rf.Network._generated_functions().keys():
        signature = signature.split(",")[2:]
        return "(" + ", ".join(signature), return_annotation

    return signature, return_annotation

def setup(app):
    app.connect("autodoc-process-signature", process_signature)



"""
.. module:: skrf.time
========================================
time (:mod:`skrf.time`)
========================================

Time domain functions


.. autosummary::
   :toctree: generated/

   time_gate
   detect_span
   find_n_peaks
   indexes
   get_window

"""
from __future__ import annotations

import warnings

# imports for type hinting
from typing import TYPE_CHECKING, Callable

import numpy as np
from numpy.fft import fft, fftshift, ifft, ifftshift, irfft, rfft
from scipy import signal
from scipy.ndimage import convolve1d

from .util import find_nearest_index

if TYPE_CHECKING:
    from .network import Network


def indexes(y: np.ndarray, thres: float = 0.3, min_dist: int = 1) -> np.ndarray:
    """
    Peak detection routine.

    Finds the numeric index of the peaks in *y* by taking its first order difference. By using
    *thres* and *min_dist* parameters, it is possible to reduce the number of
    detected peaks. *y* must be signed.

    Parameters
    ----------
    y : ndarray (signed)
        1D amplitude data to search for peaks.
    thres : float between [0., 1.], optional
        Normalized threshold. Only the peaks with amplitude higher than the
        threshold will be detected. Default is 0.3
    min_dist : int, optional
        Minimum distance between each detected peak. The peak with the highest
        amplitude is preferred to satisfy this constraint. Default is 1

    Returns
    -------
    ndarray
        Array containing the numeric indexes of the peaks that were detected

    Notes
    -----
    This function was taken from peakutils-1.1.0
    http://pythonhosted.org/PeakUtils/index.html

    """
    #This function was taken from peakutils, and is covered
    # by the MIT license, included below:

    #The MIT License (MIT)

    #Copyright (c) 2014 Lucas Hermann Negri

    #Permission is hereby granted, free of charge, to any person obtaining a copy
    #of this software and associated documentation files (the "Software"), to deal
    #in the Software without restriction, including without limitation the rights
    #to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    #copies of the Software, and to permit persons to whom the Software is
    #furnished to do so, subject to the following conditions:

    #The above copyright notice and this permission notice shall be included in
    #all copies or substantial portions of the Software.

    #THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    #IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    #FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    #AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    #LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    #OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    #THE SOFTWARE.

    if isinstance(y, np.ndarray) and np.issubdtype(y.dtype, np.unsignedinteger):
        raise ValueError("y must be signed")

    thres = thres * (np.max(y) - np.min(y)) + np.min(y)
    min_dist = int(min_dist)

    # compute first order difference
    dy = np.diff(y)

    # propagate left and right values successively to fill all plateau pixels (0-value)
    zeros, = np.where(dy == 0)

    # check if the singal is totally flat
    if len(zeros) == len(y) - 1:
        return np.array([])

    while len(zeros):
        # add pixels 2 by 2 to propagate left and right value onto the zero-value pixel
        zerosr = np.hstack([dy[1:], 0.])
        zerosl = np.hstack([0., dy[:-1]])

        # replace 0 with right value if non zero
        dy[zeros]=zerosr[zeros]
        zeros, = np.where(dy == 0)

        # replace 0 with left value if non zero
        dy[zeros] = zerosl[zeros]
        zeros, = np.where(dy == 0)

    # find the peaks by using the first order difference
    peaks = np.where((np.hstack([dy, 0.]) < 0.)
                     & (np.hstack([0., dy]) > 0.)
                     & (y > thres))[0]

    # handle multiple peaks, respecting the minimum distance
    if peaks.size > 1 and min_dist > 1:
        highest = peaks[np.argsort(y[peaks])][::-1]
        rem = np.ones(y.size, dtype=bool)
        rem[peaks] = False

        for peak in highest:
            if not rem[peak]:
                sl = slice(max(0, peak - min_dist), peak + min_dist + 1)
                rem[sl] = True
                rem[peak] = False

        peaks = np.arange(y.size)[~rem]

    return peaks


def find_n_peaks(x: np.ndarray, n: int, thres: float = 0.9, **kwargs) -> list[int]:
    """
    Find a given number of peaks in a signal.

    Parameters
    ----------
    x : np.ndarray
        signal
    n : int
        number of peaks to search for
    thres : float, optional
        threshold, default is 0.9
    **kwargs : optional keyword arguments passed to :func:`indexes`

    Returns
    -------
    peak_idxs : list of int
        List containing the numeric indexes of the peaks that were detected

    Raises
    ------
    ValueError
        If no peaks are found.
    """
    for _dummy in range(10):

        idx = indexes(x, **kwargs)
        if len(idx) < n:
            thres *= .5

        else:
            peak_vals = sorted(x[idx], reverse=True)[:n]
            peak_idxs = [x.tolist().index(k) for k in peak_vals]

            return peak_idxs
    raise ValueError('Couldnt find %i peaks' % n)

time_lookup_dict = {
    "s": 1,
    "ms": 1e-3,
    "us": 1e-6,
    "µs": 1e-6,
    "ns": 1e-9,
    "ps": 1e-12
}

def detect_span(ntwk: Network, t_unit: str = "") -> float:
    """
    Detect the correct time-span between two largest peaks.

    Parameters
    ----------
    ntwk : :class:`~skrf.network.Network`
        network to get data from

    t_unit : str
        Time unit for start, stop, center and span arguments, defaults to seconds (s).

        Possible values:
            * 's': seconds
            * 'ms': milliseconds
            * 'µs' or 'us': microseconds
            * 'ns': nanoseconds (default)
            * 'ps': picoseconds


    Returns
    -------
    span : float in unit t_unit
    """

    if t_unit == "":
        # Do not raise in autogate mode, where all parameters are None
        warnings.warn('''
                        Time unit not passed: uses 's' per default.
                        ''',
                        DeprecationWarning, stacklevel=2)
        t_unit = 's'

    x = ntwk.s_time_db.flatten()
    p1, p2 = find_n_peaks(x, n=2)
    # distance to nearest neighbor peak
    span = abs(ntwk.frequency.t[p1]-ntwk.frequency.t[p2])

    return span / time_lookup_dict[t_unit]

def get_window(window: str | tuple | Callable, Nx: int, **kwargs) -> np.ndarray:
    """Calls a custom window function or `scipy.signal.get_window()` depending on the window argument.

    Parameters
    ----------
    window : str, tuple or Callable
        The callable window function or a valid window string
        for `scipy.signal.get_window()`.
    Nx : int
        Number of samples

    Returns
    -------
    window : np.ndarray
        Window samples.
    """

    if callable(window):
        return window(Nx, **kwargs)
    else:
        return signal.get_window(window, Nx=Nx, **kwargs)

def time_gate(ntwk: Network, start: float = None, stop: float = None, center: float = None, span: float = None,
              mode: str = 'bandpass', window=('kaiser', 6),
              method: str ='fft', fft_window: str='cosine', conv_mode: str='wrap', t_unit: str = "") -> Network:
    """
    Time-domain gating of one-port s-parameters with a window function from scipy.signal.windows.

    The gate can be defined with start/stop times, or by center/span. All times are in units of seconds but
    can be changed using the `t_unit` parameter.
    Common windows are:

    * ('kaiser', 6)
    * 6 # integers are interpreted as kaiser beta-values
    * 'hamming'
    * 'boxcar'  # a straight up rect
    * Callable function which accepts integer with window size as argument

    If no parameters are passed this will try to auto-gate the largest
    peak.

    Parameters
    ----------
    ntwk : :class:`~skrf.network.Network`
        network to operate on
    start : number, or None
        start of time gate in t_unit.
    stop : number, or None
        stop of time gate in t_unit.
    center : number, or None
        center of time gate, in t_unit. If None, and span is given,
        the gate will be centered on the peak.
    span : number, or None
        span of time gate, in t_unit.  If None span will be half of the
        distance to the second tallest peak
    mode : ['bandpass', 'bandstop']
        mode of gate
    window : string, float, or tuple
        passed to `window` arg of `scipy.signal.get_window()` or callable function
    method : str
        Gating method. There are 3 option: 'convolution', 'fft', 'rfft'.

        With *'convolution'*, the time-domain gate gets transformed into frequency-domain using inverse FFT and the
        gating is then achieved by convolution with the frequency-domain data.

        With *'fft'* (default), the data gets transformed into time-domain using inverse FFT and the gating is achieved
        by multiplication with the time-domain gate. The gated time-domain signal is then transformed back into
        frequency-domain using FFT. As only positive signal frequencies are considered for the inverse FFT
        (with or without a dc component), the resulting time-domain signal has the same number of samples as in the
        frequency-domain, but is complex-valued. This method is also know as *time-domain band-pass mode*.

        With *'rfft'*, the procedure is the same as with *'fft'*, but the inverse FFT uses a complex-conjugate copy of
        the positive signal frequencies for the negative frequencies (Hermitian frequency response). A dc sample is
        also required. The resulting time-domain signal is real-valued and has twice the number of samples, which gives
        an improved time resolution. This method is also known as *time-domain low-pass mode*.

    fft_window : str or tuple or None
        Frequency-domain window applied before the inverse FFT in case of the (R)FFT method.
        This parameter takes the same values as the `window` parameter.
        Example: `window='hann` (default), or `window=('kaiser', 5)`, or `window=None`.
        The window helps to remove artefacts such as time-domain sidelobes of the pulses, but it is a trade-off with
        the achievable pulse width. The window is removed when the gated time-domain signals is transformed back into
        frequency-domain.

    conv_mode : str
        Extension mode for the convolution (if selected) determining how the frequency-domain gate is extended beyond
        the boundaries. This has a large effect on the generation of gating artefacts due to boundary effects. The
        optimal mode depends on the data. See the parameter description of `scipy.ndimage.convolve1d` for the available
        options.

    t_unit : str
        Time unit for start, stop, center and span arguments, defaults to seconds (s).

        Possible values:
            * 's': seconds
            * 'ms': milliseconds
            * 'µs' or 'us': microseconds
            * 'ns': nanoseconds (default)
            * 'ps': picoseconds


    Note
    ----
    You cant gate things that are 'behind' strong reflections. This
    is due to the multiple reflections that occur.

    If you need to time-gate an N-port network, then you should
    gate each s-parameter independently.

    Returns
    -------
    ntwk : Network
        copy of ntwk with time-gated s-parameters

    .. warning::
        Depending on sharpness of the gate, the band edges may be
        inaccurate, due to properties of FFT. We do not re-normalize
        anything.
    """

    if ntwk.nports >1:
        raise ValueError('Time-gating only works on one-ports. Try passing `ntwk.s11` or `ntwk.s21`.')

    if t_unit == "":
        if not all([e is None for e in [start, stop, center, span]]):
            # Do not raise in autogate mode, where all parameters are None
            warnings.warn('''
                            Time unit not passed: uses 's' per default.
                            ''',
                            DeprecationWarning, stacklevel=2)
        t_unit = 's'

    t_mult = time_lookup_dict[t_unit]

    if start is not None and stop is not None:
        start *= t_mult
        stop *= t_mult
        span = abs(stop-start)
        center = (stop+start)/2.

    else:
        if center is None:
            # they didnt provide center, so find the peak
            n = ntwk.s_time_mag.argmax()
            center = ntwk.frequency.t_ns[n]

        if span is None:
            span = detect_span(ntwk, t_unit='ns')

        center *= t_mult
        span *= t_mult
        start = center - span / 2.
        stop = center + span / 2.

    ntwk_gated = ntwk.copy()
    method = method.lower()
    n_fd = ntwk.frequency.npoints
    df = ntwk.frequency.step

    if method == 'convolution':
        # frequency-domain gating
        n_td = n_fd
        # create dummy-window
        window_fd = np.ones(n_fd)

    elif method == 'fft':
        # time-domain band-pass mode
        n_td = n_fd
        if fft_window is not None:
            # create band-pass window (zero on both lower and upper limit, one at center)
            window_fd = get_window(fft_window, n_fd)
        else:
            # create dummy-window
            window_fd = np.ones(n_fd)

    elif method == 'rfft':
        # time-domain low-pass mode
        if ntwk.f[0] > 0.0:
            # no dc point included
            warnings.warn('The network data to be gated does not contain the dc point (0 Hz). This is required for the '
                          'selected low-pass gating mode. Please consider to include the dc point if the results are '
                          'inaccurate, either by direct measurement of by extrapolation using '
                          'skrf.Network.extrapolate_to_dc().', UserWarning, stacklevel=2)
        n_td = 2 * n_fd - 1
        if fft_window is not None:
            # create low-pass window (one at lower limit at f=0, zero on upper limit)
            window_fd = get_window(fft_window, 2 * n_fd)
            window_fd = window_fd[n_fd:]
        else:
            # create dummy-window
            window_fd = np.ones(n_fd)

    else:
        raise ValueError(f'Invalid parameter method=`{method}`')

    # apply frequency-domain window
    ntwk_gated.s[:, 0, 0] = ntwk_gated.s[:, 0, 0] * window_fd

    # create time vector
    t = np.fft.fftshift(np.fft.fftfreq(n_td, df))
    # find start/stop gate indices
    start_idx = find_nearest_index(t, start)
    stop_idx = find_nearest_index(t, stop)

    # create gating window
    window_width = abs(stop_idx - start_idx) + 1
    window = get_window(window, window_width)

    # create the gate by padding the window with zeros
    gate = np.zeros_like(t)
    gate[start_idx:stop_idx+1] = window

    if method == 'convolution':
        # frequency-domain gating
        kernel = fftshift(fft(ifftshift(gate), norm='forward'))
        ntwk_gated.s[:, 0, 0] = convolve1d(ntwk_gated.s[:, 0, 0], kernel, mode=conv_mode)
    elif method == 'fft':
        # time-domain band-pass mode
        s_td = fftshift(ifft(ntwk_gated.s[:, 0, 0]))
        s_td_g = s_td * gate
        ntwk_gated.s[:, 0, 0] = fft(ifftshift(s_td_g))
    elif method == 'rfft':
        # time-domain low-pass mode
        s_td = fftshift(irfft(ntwk_gated.s[:, 0, 0], n=len(t)))
        s_td_g = s_td * gate
        ntwk_gated.s[:, 0, 0] = rfft(ifftshift(s_td_g))

    # remove frequency-domain window
    ntwk_gated.s[:, 0, 0] = ntwk_gated.s[:, 0, 0] / window_fd

    if mode == 'bandstop':
        ntwk_gated = ntwk - ntwk_gated
    elif mode == 'bandpass':
        pass
    else:
        raise ValueError('mode should be \'bandpass\' or \'bandstop\'')

    return ntwk_gated




import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from docscrape import ClassDoc, FunctionDoc, NumpyDocString
from docscrape_sphinx import SphinxClassDoc, SphinxDocString
from nose.tools import *

doc_txt = '''\
  numpy.multivariate_normal(mean, cov, shape=None, spam=None)

  Draw values from a multivariate normal distribution with specified
  mean and covariance.

  The multivariate normal or Gaussian distribution is a generalisation
  of the one-dimensional normal distribution to higher dimensions.

  Parameters
  ----------
  mean : (N,) ndarray
      Mean of the N-dimensional distribution.

      .. math::

         (1+2+3)/3

  cov : (N,N) ndarray
      Covariance matrix of the distribution.
  shape : tuple of ints
      Given a shape of, for example, (m,n,k), m*n*k samples are
      generated, and packed in an m-by-n-by-k arrangement.  Because
      each sample is N-dimensional, the output shape is (m,n,k,N).

  Returns
  -------
  out : ndarray
      The drawn samples, arranged according to `shape`.  If the
      shape given is (m,n,...), then the shape of `out` is is
      (m,n,...,N).

      In other words, each entry ``out[i,j,...,:]`` is an N-dimensional
      value drawn from the distribution.

  Other Parameters
  ----------------
  spam : parrot
      A parrot off its mortal coil.

  Raises
  ------
  RuntimeError
      Some error

  Warns
  -----
  RuntimeWarning
      Some warning

  Warnings
  --------
  Certain warnings apply.

  Notes
  -----

  Instead of specifying the full covariance matrix, popular
  approximations include:

    - Spherical covariance (`cov` is a multiple of the identity matrix)
    - Diagonal covariance (`cov` has non-negative elements only on the diagonal)

  This geometrical property can be seen in two dimensions by plotting
  generated data-points:

  >>> mean = [0,0]
  >>> cov = [[1,0],[0,100]] # diagonal covariance, points lie on x or y-axis

  >>> x,y = multivariate_normal(mean,cov,5000).T
  >>> plt.plot(x,y,'x'); plt.axis('equal'); plt.show()

  Note that the covariance matrix must be symmetric and non-negative
  definite.

  References
  ----------
  .. [1] A. Papoulis, "Probability, Random Variables, and Stochastic
         Processes," 3rd ed., McGraw-Hill Companies, 1991
  .. [2] R.O. Duda, P.E. Hart, and D.G. Stork, "Pattern Classification,"
         2nd ed., Wiley, 2001.

  See Also
  --------
  some, other, funcs
  otherfunc : relationship

  Examples
  --------
  >>> mean = (1,2)
  >>> cov = [[1,0],[1,0]]
  >>> x = multivariate_normal(mean,cov,(3,3))
  >>> print x.shape
  (3, 3, 2)

  The following is probably true, given that 0.6 is roughly twice the
  standard deviation:

  >>> print list( (x[0,0,:] - mean) < 0.6 )
  [True, True]

  .. index:: random
     :refguide: random;distributions, random;gauss

  '''
doc = NumpyDocString(doc_txt)


def test_signature():
    assert doc['Signature'].startswith('numpy.multivariate_normal(')
    assert doc['Signature'].endswith('spam=None)')

def test_summary():
    assert doc['Summary'][0].startswith('Draw values')
    assert doc['Summary'][-1].endswith('covariance.')

def test_extended_summary():
    assert doc['Extended Summary'][0].startswith('The multivariate normal')

def test_parameters():
    assert_equal(len(doc['Parameters']), 3)
    assert_equal([n for n,_,_ in doc['Parameters']], ['mean','cov','shape'])

    arg, arg_type, desc = doc['Parameters'][1]
    assert_equal(arg_type, '(N,N) ndarray')
    assert desc[0].startswith('Covariance matrix')
    assert doc['Parameters'][0][-1][-2] == '   (1+2+3)/3'

def test_other_parameters():
    assert_equal(len(doc['Other Parameters']), 1)
    assert_equal([n for n,_,_ in doc['Other Parameters']], ['spam'])
    arg, arg_type, desc = doc['Other Parameters'][0]
    assert_equal(arg_type, 'parrot')
    assert desc[0].startswith('A parrot off its mortal coil')

def test_returns():
    assert_equal(len(doc['Returns']), 1)
    arg, arg_type, desc = doc['Returns'][0]
    assert_equal(arg, 'out')
    assert_equal(arg_type, 'ndarray')
    assert desc[0].startswith('The drawn samples')
    assert desc[-1].endswith('distribution.')

def test_notes():
    assert doc['Notes'][0].startswith('Instead')
    assert doc['Notes'][-1].endswith('definite.')
    assert_equal(len(doc['Notes']), 17)

def test_references():
    assert doc['References'][0].startswith('..')
    assert doc['References'][-1].endswith('2001.')

def test_examples():
    assert doc['Examples'][0].startswith('>>>')
    assert doc['Examples'][-1].endswith('True]')

def test_index():
    assert_equal(doc['index']['default'], 'random')
    print(doc['index'])
    assert_equal(len(doc['index']), 2)
    assert_equal(len(doc['index']['refguide']), 2)

def non_blank_line_by_line_compare(a,b):
    a = [l for l in a.split('\n') if l.strip()]
    b = [l for l in b.split('\n') if l.strip()]
    for n,line in enumerate(a):
        if not line == b[n]:
            raise AssertionError(f"Lines {n} of a and b differ: "
                                 f"\n>>> {line}\n<<< {b[n]}\n")

def test_str():
    non_blank_line_by_line_compare(str(doc),
"""numpy.multivariate_normal(mean, cov, shape=None, spam=None)

Draw values from a multivariate normal distribution with specified
mean and covariance.

The multivariate normal or Gaussian distribution is a generalisation
of the one-dimensional normal distribution to higher dimensions.

Parameters
----------
mean : (N,) ndarray
    Mean of the N-dimensional distribution.

    .. math::

       (1+2+3)/3

cov : (N,N) ndarray
    Covariance matrix of the distribution.
shape : tuple of ints
    Given a shape of, for example, (m,n,k), m*n*k samples are
    generated, and packed in an m-by-n-by-k arrangement.  Because
    each sample is N-dimensional, the output shape is (m,n,k,N).

Returns
-------
out : ndarray
    The drawn samples, arranged according to `shape`.  If the
    shape given is (m,n,...), then the shape of `out` is is
    (m,n,...,N).

    In other words, each entry ``out[i,j,...,:]`` is an N-dimensional
    value drawn from the distribution.

Other Parameters
----------------
spam : parrot
    A parrot off its mortal coil.

Raises
------
RuntimeError :
    Some error

Warns
-----
RuntimeWarning :
    Some warning

Warnings
--------
Certain warnings apply.

See Also
--------
`some`_, `other`_, `funcs`_

`otherfunc`_
    relationship

Notes
-----
Instead of specifying the full covariance matrix, popular
approximations include:

  - Spherical covariance (`cov` is a multiple of the identity matrix)
  - Diagonal covariance (`cov` has non-negative elements only on the diagonal)

This geometrical property can be seen in two dimensions by plotting
generated data-points:

>>> mean = [0,0]
>>> cov = [[1,0],[0,100]] # diagonal covariance, points lie on x or y-axis

>>> x,y = multivariate_normal(mean,cov,5000).T
>>> plt.plot(x,y,'x'); plt.axis('equal'); plt.show()

Note that the covariance matrix must be symmetric and non-negative
definite.

References
----------
.. [1] A. Papoulis, "Probability, Random Variables, and Stochastic
       Processes," 3rd ed., McGraw-Hill Companies, 1991
.. [2] R.O. Duda, P.E. Hart, and D.G. Stork, "Pattern Classification,"
       2nd ed., Wiley, 2001.

Examples
--------
>>> mean = (1,2)
>>> cov = [[1,0],[1,0]]
>>> x = multivariate_normal(mean,cov,(3,3))
>>> print x.shape
(3, 3, 2)

The following is probably true, given that 0.6 is roughly twice the
standard deviation:

>>> print list( (x[0,0,:] - mean) < 0.6 )
[True, True]

.. index:: random
   :refguide: random;distributions, random;gauss""")


def test_sphinx_str():
    sphinx_doc = SphinxDocString(doc_txt)
    non_blank_line_by_line_compare(str(sphinx_doc),
"""
.. index:: random
   single: random;distributions, random;gauss

Draw values from a multivariate normal distribution with specified
mean and covariance.

The multivariate normal or Gaussian distribution is a generalisation
of the one-dimensional normal distribution to higher dimensions.

:Parameters:

    **mean** : (N,) ndarray

        Mean of the N-dimensional distribution.

        .. math::

           (1+2+3)/3

    **cov** : (N,N) ndarray

        Covariance matrix of the distribution.

    **shape** : tuple of ints

        Given a shape of, for example, (m,n,k), m*n*k samples are
        generated, and packed in an m-by-n-by-k arrangement.  Because
        each sample is N-dimensional, the output shape is (m,n,k,N).

:Returns:

    **out** : ndarray

        The drawn samples, arranged according to `shape`.  If the
        shape given is (m,n,...), then the shape of `out` is is
        (m,n,...,N).

        In other words, each entry ``out[i,j,...,:]`` is an N-dimensional
        value drawn from the distribution.

:Other Parameters:

    **spam** : parrot

        A parrot off its mortal coil.

:Raises:

    **RuntimeError** :

        Some error

:Warns:

    **RuntimeWarning** :

        Some warning

.. warning::

    Certain warnings apply.

.. seealso::

    :obj:`some`, :obj:`other`, :obj:`funcs`

    :obj:`otherfunc`
        relationship

.. rubric:: Notes

Instead of specifying the full covariance matrix, popular
approximations include:

  - Spherical covariance (`cov` is a multiple of the identity matrix)
  - Diagonal covariance (`cov` has non-negative elements only on the diagonal)

This geometrical property can be seen in two dimensions by plotting
generated data-points:

>>> mean = [0,0]
>>> cov = [[1,0],[0,100]] # diagonal covariance, points lie on x or y-axis

>>> x,y = multivariate_normal(mean,cov,5000).T
>>> plt.plot(x,y,'x'); plt.axis('equal'); plt.show()

Note that the covariance matrix must be symmetric and non-negative
definite.

.. rubric:: References

.. [1] A. Papoulis, "Probability, Random Variables, and Stochastic
       Processes," 3rd ed., McGraw-Hill Companies, 1991
.. [2] R.O. Duda, P.E. Hart, and D.G. Stork, "Pattern Classification,"
       2nd ed., Wiley, 2001.

.. only:: latex

   [1]_, [2]_

.. rubric:: Examples

>>> mean = (1,2)
>>> cov = [[1,0],[1,0]]
>>> x = multivariate_normal(mean,cov,(3,3))
>>> print x.shape
(3, 3, 2)

The following is probably true, given that 0.6 is roughly twice the
standard deviation:

>>> print list( (x[0,0,:] - mean) < 0.6 )
[True, True]
""")


doc2 = NumpyDocString("""
    Returns array of indices of the maximum values of along the given axis.

    Parameters
    ----------
    a : {array_like}
        Array to look in.
    axis : {None, integer}
        If None, the index is into the flattened array, otherwise along
        the specified axis""")

def test_parameters_without_extended_description():
    assert_equal(len(doc2['Parameters']), 2)

doc3 = NumpyDocString("""
    my_signature(*params, **kwds)

    Return this and that.
    """)

def test_escape_stars():
    signature = str(doc3).split('\n')[0]
    assert_equal(signature, r'my_signature(\*params, \*\*kwds)')

doc4 = NumpyDocString(
    """a.conj()

    Return an array with all complex-valued elements conjugated.""")

def test_empty_extended_summary():
    assert_equal(doc4['Extended Summary'], [])

doc5 = NumpyDocString(
    """
    a.something()

    Raises
    ------
    LinAlgException
        If array is singular.

    Warns
    -----
    SomeWarning
        If needed
    """)

def test_raises():
    assert_equal(len(doc5['Raises']), 1)
    name,_,desc = doc5['Raises'][0]
    assert_equal(name,'LinAlgException')
    assert_equal(desc,['If array is singular.'])

def test_warns():
    assert_equal(len(doc5['Warns']), 1)
    name,_,desc = doc5['Warns'][0]
    assert_equal(name,'SomeWarning')
    assert_equal(desc,['If needed'])

def test_see_also():
    doc6 = NumpyDocString(
    """
    z(x,theta)

    See Also
    --------
    func_a, func_b, func_c
    func_d : some equivalent func
    foo.func_e : some other func over
             multiple lines
    func_f, func_g, :meth:`func_h`, func_j,
    func_k
    :obj:`baz.obj_q`
    :class:`class_j`: fubar
        foobar
    """)

    assert len(doc6['See Also']) == 12
    for func, desc, role in doc6['See Also']:
        if func in ('func_a', 'func_b', 'func_c', 'func_f',
                    'func_g', 'func_h', 'func_j', 'func_k', 'baz.obj_q'):
            assert(not desc)
        else:
            assert(desc)

        if func == 'func_h':
            assert role == 'meth'
        elif func == 'baz.obj_q':
            assert role == 'obj'
        elif func == 'class_j':
            assert role == 'class'
        else:
            assert role is None

        if func == 'func_d':
            assert desc == ['some equivalent func']
        elif func == 'foo.func_e':
            assert desc == ['some other func over', 'multiple lines']
        elif func == 'class_j':
            assert desc == ['fubar', 'foobar']

def test_see_also_print():
    class Dummy:
        """
        See Also
        --------
        func_a, func_b
        func_c : some relationship
                 goes here
        func_d
        """
        pass

    obj = Dummy()
    s = str(FunctionDoc(obj, role='func'))
    assert(':func:`func_a`, :func:`func_b`' in s)
    assert('    some relationship' in s)
    assert(':func:`func_d`' in s)

doc7 = NumpyDocString("""

        Doc starts on second line.

        """)

def test_empty_first_line():
    assert doc7['Summary'][0].startswith('Doc starts')


def test_no_summary():
    str(SphinxDocString("""
    Parameters
    ----------"""))


def test_unicode():
    doc = SphinxDocString("""
    öäöäöäöäöåååå

    öäöäöäööäååå

    Parameters
    ----------
    ååå : äää
        ööö

    Returns
    -------
    ååå : ööö
        äää

    """)
    assert doc['Summary'][0] == 'öäöäöäöäöåååå'.encode()

def test_plot_examples():
    cfg = dict(use_plots=True)

    doc = SphinxDocString("""
    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> plt.plot([1,2,3],[4,5,6])
    >>> plt.show()
    """, config=cfg)
    assert 'plot::' in str(doc), str(doc)

    doc = SphinxDocString("""
    Examples
    --------
    .. plot::

       import matplotlib.pyplot as plt
       plt.plot([1,2,3],[4,5,6])
       plt.show()
    """, config=cfg)
    assert str(doc).count('plot::') == 1, str(doc)

def test_class_members():

    class Dummy:
        """
        Dummy class.

        """
        def spam(self, a, b):
            """Spam\n\nSpam spam."""
            pass
        def ham(self, c, d):
            """Cheese\n\nNo cheese."""
            pass

    for cls in (ClassDoc, SphinxClassDoc):
        doc = cls(Dummy, config=dict(show_class_members=False))
        assert 'Methods' not in str(doc), (cls, str(doc))
        assert 'spam' not in str(doc), (cls, str(doc))
        assert 'ham' not in str(doc), (cls, str(doc))

        doc = cls(Dummy, config=dict(show_class_members=True))
        assert 'Methods' in str(doc), (cls, str(doc))
        assert 'spam' in str(doc), (cls, str(doc))
        assert 'ham' in str(doc), (cls, str(doc))

        if cls is SphinxClassDoc:
            assert '.. autosummary::' in str(doc), str(doc)

if __name__ == "__main__":
    import nose
    nose.run()



"""

.. currentmodule:: skrf.instances
========================================
instances (:mod:`skrf.instances`)
========================================

This module contains commonly used  instantiated objects's.



Standard Waveguide Bands
----------------------------

:class:`~skrf.frequency.Frequency` Objects
++++++++++++++++++++++++++++++++++++++++++++
These are predefined :class:`~skrf.frequency.Frequency` objects
that correspond to standard waveguide bands. This information is taken
from the VDI Application Note 1002 [#]_ . IEEE designators are taken
from Spinner TD-00036 [#]_ .


=======================  ===============================================
Object Name              Description
=======================  ===============================================
f_wr10                   WR-10, 75-110 GHz
f_wr3                    WR-3, 220-325 GHz
f_wr2p2                  WR-2.2, 330-500 GHz
f_wr1p5                  WR-1.5, 500-750 GHz
f_wr1                    WR-1, 750-1100 GHz
...                      ...
=======================  ===============================================


:class:`~skrf.media.rectangularWaveguide.RectangularWaveguide`  Objects
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
These are predefined :class:`~skrf.media.rectangularWaveguide.RectangularWaveguide`
objects for  standard waveguide bands.

=======================  ===============================================
Object Name              Description
=======================  ===============================================
wr10                     WR-10, 75-110 GHz
wr3                      WR-3, 220-325 GHz
wr2p2                    WR-2.2, 330-500 GHz
wr1p5                    WR-1.5, 500-750 GHz
wr1                      WR-1, 750-1100 GHz
...                      ...
=======================  ===============================================



References
----------
.. [#] VDI Application Note:  VDI Waveguide Band Designations (VDI-1002) http://vadiodes.com/VDI/pdf/waveguidechart200908.pdf
.. [#] Spinner Technical Information: Cross Reference For Hollow Metallic Waveguides (TD-00036) https://www.spinner-group.com/images/download/technical_documents/SPINNER_TD00036.pdf
"""

from .constants import mil
from .frequency import Frequency
from .media import Freespace, RectangularWaveguide


class StaticInstances:
    @property
    def air(self) -> Freespace:
        return Freespace()

    @property
    def air50(self) -> Freespace:
        return Freespace(z0_override=50)

    @property
    def f_wr51(self) -> Frequency:
        return Frequency(15,22,1001, 'GHz')

    @property
    def f_wr42(self) -> Frequency:
        return Frequency(17.5,26.5,1001, 'GHz')

    @property
    def f_wr34(self) -> Frequency:
        return Frequency(22,33,1001, 'GHz')

    @property
    def f_wr28(self) -> Frequency:
        return Frequency(26.5,40,1001, 'GHz')

    @property
    def f_wr22p4(self) -> Frequency:
        return Frequency(33,50.5,1001, 'GHz')

    @property
    def f_wr18p8(self) -> Frequency:
        return Frequency(40,60,1001, 'GHz')

    @property
    def f_wr14p8(self) -> Frequency:
        return Frequency(50,75,1001, 'GHz')

    @property
    def f_wr12p2(self) -> Frequency:
        return Frequency(60,90,1001, 'GHz')

    @property
    def f_wr10(self) -> Frequency:
        return Frequency(75,110,1001, 'GHz')

    @property
    def f_wr8(self) -> Frequency:
        return Frequency(90,140,1001, 'GHz')

    @property
    def f_wr6p5(self) -> Frequency:
        return Frequency(110,170,1001, 'GHz')

    @property
    def f_wr5p1(self) -> Frequency:
        return Frequency(140,220,1001, 'GHz')

    @property
    def f_wr4p3(self) -> Frequency:
        return Frequency(170,260,1001, 'GHz')

    @property
    def f_wr3p4(self) -> Frequency:
        return Frequency(220,330,1001, 'GHz')

    @property
    def f_wr2p8(self) -> Frequency:
        return Frequency(260,400,1001, 'GHz')

    @property
    def f_wr2p2(self) -> Frequency:
        return Frequency(330,500,1001, 'GHz')

    @property
    def f_wr1p9(self) -> Frequency:
        return Frequency(400,600,1001, 'GHz')

    @property
    def f_wr1p5(self) -> Frequency:
        return Frequency(500,750,1001, 'GHz')

    @property
    def f_wr1p2(self) -> Frequency:
        return Frequency(600,900,1001, 'GHz')

    @property
    def f_wr1(self) -> Frequency:
        return Frequency(750,1100,1001, 'GHz')

    @property
    def f_wr0p8(self) -> Frequency:
        return Frequency(900,1400,1001, 'GHz')

    @property
    def f_wr0p65(self) -> Frequency:
        return Frequency(1100,1700,1001, 'GHz')

    @property
    def f_wr0p51(self) -> Frequency:
        return Frequency(1400,2200,1001, 'GHz')

    @property
    def f_wm106(self) -> Frequency:
        return Frequency(1700,2600,1001, 'ghz')

    @property
    def f_wm86(self) -> Frequency:
        return Frequency(2200,3300,1001, 'ghz')

    @property
    def f_wm1295(self) -> Frequency:
        return self.f_wr5p1

    @property
    def f_wm1092(self) -> Frequency:
        return self.f_wr4p3

    @property
    def f_wm864(self) -> Frequency:
        return self.f_wr3p4

    @property
    def f_wm710(self) -> Frequency:
        return self.f_wr2p8

    @property
    def f_wm570(self) -> Frequency:
        return self.f_wr2p2

    @property
    def f_wm470(self) -> Frequency:
        return self.f_wr1p9

    @property
    def f_wm380(self) -> Frequency:
        return self.f_wr1p5

    @property
    def f_wm310(self) -> Frequency:
        return self.f_wr1p2

    @property
    def f_wm250(self) -> Frequency:
        return self.f_wr1

    @property
    def f_wm200(self) -> Frequency:
        return self.f_wr0p8

    @property
    def f_wm164(self) -> Frequency:
        return self.f_wr0p65

    @property
    def f_wm130(self) -> Frequency:
        return self.f_wr0p51

    @property
    def wr51(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr51.copy(),a=510*mil,b=255*mil,z0_override=50)

    @property
    def wr42(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr42.copy(),a=420*mil,b=170*mil,z0_override=50)

    @property
    def wr34(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr34.copy(),a=340*mil,b=170*mil,z0_override=50)

    @property
    def wr28(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr28.copy(),a=280*mil,b=140*mil,z0_override=50)

    @property
    def wr22p4(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr22p4.copy(),a=224*mil,b=112*mil,z0_override=50)

    @property
    def wr18p8(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr18p8.copy(),a=188*mil,b=94*mil,z0_override=50)

    @property
    def wr14p8(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr14p8.copy(),a=148*mil,b=74*mil,z0_override=50)

    @property
    def wr12p2(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr12p2.copy(),a=122*mil,b=61*mil,z0_override=50)

    @property
    def wr10(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr10.copy(),a=100*mil,b=50*mil,z0_override=50)

    @property
    def wr8(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr8.copy(),a=80*mil,b=40*mil,z0_override=50)

    @property
    def wr6p5(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr6p5.copy(),a=65*mil,b=32.5*mil,z0_override=50)

    @property
    def wr5p1(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr5p1.copy(),a=51*mil,b=25.5*mil,z0_override=50)

    @property
    def wr4p3(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr4p3.copy(),a=43*mil,b=21.5*mil,z0_override=50)

    @property
    def wr3p4(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr3p4.copy(),a=34*mil,b=17*mil,z0_override=50)

    @property
    def wr2p8(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr2p8.copy(),a=28*mil,b=14*mil,z0_override=50)

    @property
    def wr2p2(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr2p2.copy(),a=22*mil,b=11*mil,z0_override=50)

    @property
    def wr1p9(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr1p9.copy(),a=19*mil,b=9.5*mil,z0_override=50)

    @property
    def wr1p5(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr1p5.copy(),a=15*mil,b=7.5*mil,z0_override=50)

    @property
    def wr1p2(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr1p2.copy(),a=12*mil,b=6*mil,z0_override=50)

    @property
    def wr1(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr1.copy(),a=10*mil,b=5*mil,z0_override=50)

    @property
    def wr0p8(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr0p8.copy(),a=8*mil,b=4*mil,z0_override=50)

    @property
    def wr0p65(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr0p65.copy(),a=6.5*mil,b=3.25*mil,z0_override=50)

    @property
    def wr0p51(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wr0p51.copy(),a=5.1*mil,b=2.55*mil,z0_override=50)

    @property
    def wm1295(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wm1295.copy(),a=1295e-6,b=647.5e-6,z0_override=50)

    @property
    def wm1092(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wm1092.copy(),a=1092e-6,b=546e-6,z0_override=50)

    @property
    def wm864(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wm864.copy(),a=864e-6,b=432e-6,z0_override=50)

    @property
    def wm710(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wm710.copy(),a=710e-6,b=355e-6,z0_override=50)

    @property
    def wm570(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wm570.copy(),a=570e-6,b=285e-6,z0_override=50)

    @property
    def wm470(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wm470.copy(),a=470e-6,b=235e-6,z0_override=50)

    @property
    def wm380(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wm380.copy(),a=380e-6,b=190e-6,z0_override=50)

    @property
    def wm310(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wm310.copy(),a=310e-6,b=155e-6,z0_override=50)

    @property
    def wm250(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wm250.copy(),a=250e-6,b=125e-6,z0_override=50)

    @property
    def wm200(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wm200.copy(),a=200e-6,b=100e-6,z0_override=50)

    @property
    def wm164(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wm164.copy(),a=164e-6,b=82e-6,z0_override=50)

    @property
    def wm130(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wm130.copy(),a=130e-6,b=65e-6,z0_override=50)

    @property
    def wm106(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wm106.copy(),a=106e-6,b=53e-6,z0_override=50)

    @property
    def wm86(self) -> RectangularWaveguide:
        return RectangularWaveguide(self.f_wm86.copy(),a=86e-6,b=43e-6,z0_override=50)

_instances = StaticInstances()

def __getattr__(name):
    return getattr(_instances, name)




import sys

import numpy as np
import pytest

import skrf

try:
    from skrf.vi.vna import ValuesFormat, keysight
    from skrf.vi.vna.keysight.pna import SweepMode, SweepType
except ImportError:
    pass

if "matplotlib" not in sys.modules:
    pytest.skip(allow_module_level=True)

@pytest.fixture
def mocked_ff(mocker):
    mocker.patch('skrf.vi.vna.keysight.PNA.__init__', return_value=None)
    mocker.patch('skrf.vi.vna.keysight.PNA.write')
    mocker.patch('skrf.vi.vna.keysight.PNA.write_values')
    mocker.patch('skrf.vi.vna.keysight.PNA.query')
    mocker.patch('skrf.vi.vna.keysight.PNA.query_values')
    mock = keysight.PNA('TEST')
    mock.model = "TEST"

    # This gets done in init, but we are mocking init to prevent super().__init__, so just call here
    mock.create_channel(1, 'Channel 1')

    yield mock

@pytest.mark.parametrize(
    'param,expected_query,expected_write,query_response,expected_val,write_val',
    [
        ('freq_start', 'SENS1:FREQ:STAR?', 'SENS1:FREQ:STAR 100', '100', 100, 100),
        ('freq_stop', 'SENS1:FREQ:STOP?', 'SENS1:FREQ:STOP 100', '100', 100, 100),
        ('freq_span', 'SENS1:FREQ:SPAN?', 'SENS1:FREQ:SPAN 100', '100', 100, 100),
        ('freq_center', 'SENS1:FREQ:CENT?', 'SENS1:FREQ:CENT 100', '100', 100, 100),
        ('npoints', 'SENS1:SWE:POIN?', 'SENS1:SWE:POIN 100', '100', 100, 100),
        ('if_bandwidth', 'SENS1:BWID?', 'SENS1:BWID 100', '100', 100, 100),
        ('sweep_time', 'SENS1:SWE:TIME?', 'SENS1:SWE:TIME 1.0', '1.0', 1.0, 1),
        ('sweep_type', 'SENS1:SWE:TYPE?', 'SENS1:SWE:TYPE LIN', 'LIN', SweepType.LINEAR, SweepType.LINEAR),
        ('sweep_mode', 'SENS1:SWE:MODE?', 'SENS1:SWE:MODE SING', 'SING', SweepMode.SINGLE, SweepMode.SINGLE),
        ('measurement_numbers', 'SYST:MEAS:CAT? 1', None, '1,2,3', [1, 2, 3], None),
    ]
)
def test_params(
    mocker,
    mocked_ff,
    param,
    expected_query,
    expected_write,
    query_response,
    expected_val,
    write_val
):
    if expected_query is not None:
        mocked_ff.query.return_value = query_response
        test_query = getattr(mocked_ff.ch1, param)
        mocked_ff.query.assert_called_once_with(expected_query)
        assert test_query == expected_val

    if expected_write is not None:
        setattr(mocked_ff.ch1, param, write_val)
        mocked_ff.write.assert_called_once_with(expected_write)

def test_frequency_query(mocker, mocked_ff):
    mocked_ff.query.side_effect = [
        '100', '200', '11'
    ]
    test = mocked_ff.ch1.frequency
    assert test == skrf.Frequency(100, 200, 11, unit='hz')

def test_frequency_write(mocker, mocked_ff):
    test_f = skrf.Frequency(100, 200, 11, unit='hz')
    mocked_ff.ch1.frequency = test_f
    calls = [
        mocker.call("SENS1:FREQ:STAR 100"),
        mocker.call("SENS1:FREQ:STOP 200"),
        mocker.call("SENS1:SWE:POIN 11"),
    ]
    mocked_ff.write.assert_has_calls(calls)

# def test_create_channel(mocker, mocked_ff):
    # mocked_ff.create_channel(2, 'Channel 2')
    # assert hasattr(mocked_ff, 'ch2')
    # assert mocked_ff.ch2.cnum == 2
    # assert mocked_ff.ch2.name == "Channel 2"

def test_active_channel_query(mocker, mocked_ff):
    mocked_ff.query.return_value = 1
    test = mocked_ff.active_channel
    assert isinstance(test, keysight.PNA.Channel)
    assert test.cnum == 1

def test_active_channel_setter(mocker, mocked_ff):
    mocked_ff.query.side_effect = ['1', '1', '1', '1,2,3', '2']
    mocked_ff.active_channel = mocked_ff.ch1
    mocked_ff.write.assert_not_called()

    mocked_ff.create_channel(2, 'Test')
    mocked_ff.active_channel = mocked_ff.ch2

    assert mocked_ff.active_channel.cnum == 2

def test_query_fmt_query(mocker, mocked_ff):
    mocked_ff.query.side_effect = ['ASC,0', 'REAL,32', 'REAL,64']
    test = mocked_ff.query_format
    assert test == ValuesFormat.ASCII
    test = mocked_ff.query_format
    assert test == ValuesFormat.BINARY_32
    test = mocked_ff.query_format
    assert test == ValuesFormat.BINARY_64

def test_query_fmt_write(mocker, mocked_ff):
    mocked_ff.query_format = ValuesFormat.ASCII
    mocked_ff.write.assert_called_with('FORM ASC,0')
    mocked_ff.query_format = ValuesFormat.BINARY_32
    calls = [
        mocker.call("FORM:BORD SWAP"),
        mocker.call("FORM REAL,32"),
    ]
    mocked_ff.write.assert_has_calls(calls)
    mocked_ff.query_format = ValuesFormat.BINARY_64
    calls = [
        mocker.call("FORM:BORD SWAP"),
        mocker.call("FORM REAL,64"),
    ]
    mocked_ff.write.assert_has_calls(calls)

def test_measurements_query(mocker, mocked_ff):
    mocked_ff.query.return_value = 'CH1_S11_1,S11,CH1_S12_1,S12'
    test = mocked_ff.ch1.measurements
    assert test == [('CH1_S11_1', 'S11'), ('CH1_S12_1', 'S12')]

def test_measurement_names_query(mocker, mocked_ff):
    mocked_ff.query.return_value = 'CH1_S11_1,S11,CH1_S12_1,S12'
    test = mocked_ff.ch1.measurement_names
    assert test == ['CH1_S11_1', 'CH1_S12_1']

def test_clear_averaging(mocker, mocked_ff):
    mocked_ff.ch1.clear_averaging()
    mocked_ff.write.assert_called_once_with('SENS1:AVER:CLE')

def test_create_measurement(mocker, mocked_ff):
    mocked_ff.query.return_value = '1'
    mocked_ff.ch1.create_measurement('CH1_S11_1', 'S11')
    write_calls = [
        mocker.call("CALC1:PAR:EXT 'CH1_S11_1',S11"),
        mocker.call("DISP:WIND:TRAC2:FEED 'CH1_S11_1'"),
    ]

    mocked_ff.write.assert_has_calls(write_calls)

def test_delete_measurement(mocker, mocked_ff):
    mocked_ff.ch1.delete_measurement('CH1_S11_1')
    mocked_ff.write.assert_called_once_with("CALC1:PAR:DEL 'CH1_S11_1'")

def test_get_measurement(mocker, mocked_ff):
    mocked_ff.ch1.get_active_trace = mocker.MagicMock(return_value=skrf.Network())
    mocked_ff.query.side_effect = [
        'CH1_S11_1,S11,CH1_S12_1,S12',
        'CH1_S11_1,S11,CH1_S12_1,S12',
    ]
    test = mocked_ff.ch1.get_measurement('CH1_S11_1')
    mocked_ff.write.assert_called_once_with("CALC1:PAR:SEL 'CH1_S11_1',fast")
    assert isinstance(test, skrf.Network)

def test_get_active_trace(mocker, mocked_ff):
    mock_sdata = np.array([1.,]*22)
    query_responses = [
        'ASC,0',
        '100','200','11'
    ]
    expected_writes = [
        mocker.call('FORM:BORD SWAP'),
        mocker.call('FORM REAL,64'),
        mocker.call('FORM ASC,0')
    ]
    mocked_ff.query.side_effect = query_responses
    mocked_ff.ch1.sweep = mocker.MagicMock()
    mocker.patch('skrf.vi.vna.keysight.PNA.Channel.active_trace_sdata',
                return_value=mock_sdata, new_callable=mocker.PropertyMock)
    test = mocked_ff.ch1.get_active_trace()
    assert isinstance(test, skrf.Network)
    mocked_ff.ch1.sweep.assert_called_once()
    mocked_ff.write.assert_has_calls(expected_writes)

# Getting the query responses right is annoying. A lot of checks and queries
# happen when getting the snp network, and specifying them here is difficult
# and error-prone especially when changes are made. Something to figure out
# for the future
# def test_get_snp_network(mocker, mocked_ff):
#     mock_sdata = np.array([1.,]*22*4)
#     query_responses = [
#         '4',
#         'ASC,0',
#         '1', '1', '1', '1', '1', '1', '1', '1', '1',
#         '1', '1', '1', '1', '1', '1', '1', '1',
#         '11',
#         'TEST1', '1',
#         'TEST2', '2',
#         'TEST3', '3',
#         'TEST4', '4',
#         '1',
#         '100','200','11'
#     ]
#     expected_writes = [
#         mocker.call('FORM:BORD SWAP;FORM REAL,64'),
#         mocker.call("CALC1:PAR:EXT 'TEST1',S11"),
#         mocker.call("DISP:WIND:TRAC:FEED1 'TEST1'"),
#         mocker.call("CALC1:PAR:EXT 'TEST2',S22"),
#         mocker.call("DISP:WIND:TRAC:FEED2 'TEST2'"),
#         mocker.call("CALC1:PAR:EXT 'TEST3',S33"),
#         mocker.call("DISP:WIND:TRAC:FEED3 'TEST3'"),
#         mocker.call("CALC1:PAR:EXT 'TEST4',S44"),
#         mocker.call("DISP:WIND:TRAC:FEED4 'TEST4'"),
#         mocker.call("CALC1:PAR:DEL 'TEST1'"),
#         mocker.call("CALC1:PAR:DEL 'TEST2'"),
#         mocker.call("CALC1:PAR:DEL 'TEST3'"),
#         mocker.call("CALC1:PAR:DEL 'TEST4'"),
#         mocker.call('FORM ASC,0')
#     ]
#     mocked_ff.wait_for_complete = mocker.MagicMock()
#     mocked_ff.query.side_effect = query_responses
#     mocked_ff.ch1.sweep = mocker.MagicMock()
#     mocked_ff.query_values.return_value=mock_sdata
#     test = mocked_ff.ch1.get_snp_network()
#     assert isinstance(test, skrf.Network)
#     mocked_ff.ch1.sweep.assert_called_once()
#     mocked_ff.write.assert_has_calls(expected_writes)



"""
.. module:: skrf.vi.vna.vna
=================================================
vna (:mod:`skrf.vi.vna.vna`)
=================================================

Provides the VNA base class
"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    pass

import functools
import inspect
import re
from enum import Enum, auto

import numpy as np
import pyvisa

from ..scpi_errors import SCPIError
from ..validators import Validator


def _format_cmd(cmd: str, **kwargs) -> str:
    def sub(match_obj):
        prefix = match_obj.group("prefix")
        attr = match_obj.group("attr")

        if prefix:
            obj = kwargs[prefix]
            return str(getattr(obj, attr))
        else:
            return str(kwargs[attr])

    param_re = re.compile(r"\<(?:(?P<prefix>\w+):)?(?P<attr>\w+)\>")
    return re.sub(param_re, sub, cmd)


class ValuesFormat(Enum):
    """How values are written to and queried from the insturment"""

    #: 32 bits per value
    BINARY_32 = auto()
    #: 64 bits per value
    BINARY_64 = auto()
    #: Transferred as ASCII (e.g. representing numbers as strings)
    ASCII = auto()


class Channel:
    """
    A single channel of the instrument.

    This is only for those instruments which support channels, and should be
    subclassed in those instrument classes.

    .. warning::
        This class should not be instantiated directly
    """

    def __init__(
        self, parent, cnum: int | None = None, cname: str | None = None
    ) -> None:
        self.parent = parent
        self.cnum = cnum
        self.name = cname

        self.read = self.parent.read
        self.read_values = self.parent.read_values
        self.write = self.parent.write
        self.write_values = self.parent.write_values
        self.query = self.parent.query
        self.query_values = self.parent.query_values


class VNA:
    _scpi = True  # Set to false in subclasses that don't use SCPI

    def __init__(
        self, address: str, backend: str = "@py", timeout: int | None = None
    ) -> None:
        rm = pyvisa.ResourceManager(backend)
        self._resource = rm.open_resource(address, timeout=timeout)

        # Reading and setting the query values format is instrument specific
        # and must be done for each subclass. We default to using ASCII
        self._values_fmt: ValuesFormat = ValuesFormat.ASCII

        if self._scpi:
            self._setup_scpi()

        self.echo = False

    def __init_subclass__(cls):
        if "Channel" in [c[0] for c in inspect.getmembers(cls, inspect.isclass)]:
            cls._add_channel_support()

    @classmethod
    def _add_channel_support(cls):
        def create_channel(self, cnum: int, cname: str) -> None:
            ch_id = f"ch{cnum}"
            if hasattr(self, ch_id):
                raise RuntimeError(f"Channel {cnum} already exists")

            new_channel = self.Channel(self, cnum, cname)
            setattr(self, ch_id, new_channel)

        def delete_channel(self, cnum: str) -> None:
            ch_id = f"ch{cnum}"
            if not hasattr(self, ch_id):
                return
            ch = getattr(self, ch_id)
            if hasattr(ch, "_on_delete"):
                ch._on_delete()
            delattr(self, ch_id)

        def _channels(self) -> list[Channel]:
            return [getattr(self, ch) for ch in dir(self) if re.fullmatch(r"ch\d+", ch)]

        def __getattr__(self, k):
            if not hasattr(self.Channel, k):
                raise AttributeError(f"{type(self).__name__} has no attribute {k}")
            return getattr(self.ch1, k)

        cls.create_channel = create_channel
        cls.delete_channel = delete_channel
        cls.channels = property(_channels)
        cls.__getattr__ = __getattr__

    def _setup_scpi(self) -> None:
        self.__class__.wait_for_complete = lambda self: self.query("*OPC?")
        self.__class__.status = property(lambda self: self.query("*STB?"))
        self.__class__.options = property(lambda self: self.query("*OPT?"))
        self.__class__.id = property(lambda self: self.query("*IDN?"))
        self.__class__.clear_errors = lambda self: self.write("*CLS")

        def errcheck(self) -> None:
            err = self.query("SYST:ERR?")
            errno = int(err.split(",")[0])
            if errno == 0:
                return
            else:
                raise SCPIError(errno)

        self.__class__.check_errors = errcheck

    @staticmethod
    def command(
        get_cmd: str | None = None,
        set_cmd: str | None = None,
        doc: str | None = None,
        validator: Validator | None = None,
        values: bool = False,
        values_container: type | None = np.array,
        complex_values: bool = False,
    ) -> property:
        """
        Create a property for the instrument.

        This method is used to add a property to an instrument. These properties
        can be read-only, write-only, read-write, and can validate values before
        sending to the instrument as well as validate responses from the
        instrument to return proper types.

        Parameters
        ----------
        get_cmd
            Command sent to the instrument to request data
        set_cmd
            Command sent to the instrument to set data
        doc
            The docstring for the property
        validator
            The :class:`Validator` that will be used to transform data to the
            proper format before sending and after querying
        values
            Whether or not this command is using a `Sequence` to set data, or
            expects a `Sequence` in response.
        values_container:
            If values is true, you set set this to the type of container the
            values should be returned in. For example, this is np.array by
            default, meaning instead of return a `list`, you will get a numpy
            array.
        complex_values:
            If the values expected from the instrument are complex. If so, the
            values will be converted from [real[0], imag[0], real[1], imag[1], ...]
            to [complex(real[0], imag[0]), complex(real[1], imag[1]), ...]

        Returns
        -------
        property
            The property constructed from the parameters passed. Should be set
            to a class variable
        """

        def fget(self, get_cmd=get_cmd, validator=validator):
            if get_cmd is None:
                raise LookupError("Property cannot be read")

            cmd = _format_cmd(get_cmd, self=self)
            if values:
                arg = self.query_values(
                    cmd, container=values_container, complex_values=complex_values
                )
            else:
                arg = self.query(cmd)

            if hasattr(self, "wait_for_complete"):
                self.wait_for_complete()

            if validator:
                return validator.validate_output(arg)
            else:
                return arg

        def fset(self, arg, set_cmd=set_cmd, validator=validator):
            if set_cmd is None:
                raise LookupError("Property cannot be set")

            if validator:
                arg = validator.validate_input(arg)

            cmd = _format_cmd(set_cmd, self=self, arg=arg)
            self.write(cmd)

            if hasattr(self, "wait_for_complete"):
                self.wait_for_complete()

        fget.__doc__ = doc
        # TODO: Potentially add the validator docstring to add the verbosity to
        # the generated docs, but keep the code less verbose?

        return property(fget=fget, fset=fset)

    @property
    def timeout(self) -> int | None:
        return self._resource.timeout

    @timeout.setter
    def timeout(self, timeout: int | None) -> None:
        self._resource.timeout = timeout

    def read(self, **kwargs) -> None:
        if isinstance(self._resource, pyvisa.resources.MessageBasedResource):
            fn = self._resource.read
        elif isinstance(self._resource, pyvisa.resources.RegisterBasedResource):
            raise NotImplementedError()
        else:
            raise RuntimeError("unreachable")

        return fn(**kwargs)

    def read_values(self, **kwargs) -> None:  # noqa: B027
        pass

    def write(self, cmd, **kwargs) -> None:
        if self.echo:
            print(cmd)

        if isinstance(self._resource, pyvisa.resources.MessageBasedResource):
            fn = self._resource.write
        elif isinstance(self._resource, pyvisa.resources.RegisterBasedResource):
            raise NotImplementedError()
        else:
            raise RuntimeError("unreachable")

        fn(cmd, **kwargs)

    def write_values(self, cmd, values, complex_values: bool = False, **kwargs) -> None:
        if self.echo:
            print(cmd)

        if complex_values:
            values = np.array([(x.real, x.imag) for x in values]).flatten()

        if isinstance(self._resource, pyvisa.resources.MessageBasedResource):
            if self._values_fmt == ValuesFormat.ASCII:
                fn = self._resource.write_ascii_values
            elif self._values_fmt == ValuesFormat.BINARY_32:
                fn = self._resource.write_binary_values
            elif self._values_fmt == ValuesFormat.BINARY_64:
                fn = functools.partial(self._resource.write_binary_values, datatype="d")

        elif isinstance(self._resource, pyvisa.resources.RegisterBasedResource):
            raise NotImplementedError()
        else:
            raise RuntimeError("unreachable")

        return fn(cmd, values, **kwargs)

    def query(self, cmd, **kwargs) -> None:
        if self.echo:
            print(cmd)

        if isinstance(self._resource, pyvisa.resources.MessageBasedResource):
            fn = self._resource.query
        elif isinstance(self._resource, pyvisa.resources.RegisterBasedResource):
            raise NotImplementedError()
        else:
            raise RuntimeError("unreachable")

        return fn(cmd, **kwargs)

    def query_values(self, cmd, complex_values: bool = False, **kwargs) -> None:
        if self.echo:
            print(cmd)

        if isinstance(self._resource, pyvisa.resources.MessageBasedResource):
            if self._values_fmt == ValuesFormat.ASCII:
                fn = self._resource.query_ascii_values
            elif self._values_fmt == ValuesFormat.BINARY_32:
                fn = self._resource.query_binary_values
            elif self._values_fmt == ValuesFormat.BINARY_64:
                fn = functools.partial(self._resource.query_binary_values, datatype="d")
        elif isinstance(self._resource, pyvisa.resources.RegisterBasedResource):
            raise NotImplementedError()
        else:
            raise RuntimeError("unreachable")

        vals = fn(cmd, **kwargs)

        if complex_values:
            vals = np.vectorize(complex)(vals[::2], vals[1::2])

        return vals



"""
circuit (:mod:`skrf.circuit`)
========================================

The Circuit class represents a circuit of arbitrary topology,
consisting of an arbitrary number of N-ports networks.

Like in an electronic circuit simulator, the circuit must have one or more ports
connected to the circuit. The Circuit object allows one retrieving the M-ports network,
where M is the number of ports defined.

The results are returned in :class:`~skrf.circuit.Circuit` object.


Building a Circuit
------------------
.. autosummary::
   :toctree: generated/

   Circuit
   Circuit.Port
   Circuit.SeriesImpedance
   Circuit.ShuntAdmittance
   Circuit.Ground
   Circuit.Open
   Circuit.update_networks

Representing a Circuit
----------------------
.. autosummary::
   :toctree: generated/

   Circuit.plot_graph

Network Representations
-----------------------
.. autosummary::
   :toctree: generated/

   Circuit.network
   Circuit.s
   Circuit.s_external
   Circuit.s_active
   Circuit.z_active
   Circuit.y_active
   Circuit.vswr_active
   Circuit.port_z0

Voltages and Currents
---------------------
.. autosummary::
   :toctree: generated/

   Circuit.voltages
   Circuit.voltages_external
   Circuit.currents
   Circuit.currents_external

Circuit internals
------------------
.. autosummary::
   :toctree: generated/

   Circuit.networks_dict
   Circuit.networks_list
   Circuit.connections_nb
   Circuit.connections_list
   Circuit.nodes_nb
   Circuit.dim
   Circuit.intersections_dict
   Circuit.port_indexes
   Circuit.C
   Circuit.X

Circuit reduction
------------------
.. autosummary::
   :toctree: generated/

   reduce_circuit

Graph representation
--------------------
.. autosummary::
   :toctree: generated/

   Circuit.graph
   Circuit.G
   Circuit.edges
   Circuit.edge_labels

"""
from __future__ import annotations

from functools import cached_property
from itertools import chain
from typing import TYPE_CHECKING, Sequence, TypedDict

import numpy as np
from typing_extensions import NotRequired, Unpack

from .constants import S_DEF_DEFAULT, MemoryLayoutT, NumberLike
from .media import media
from .network import Network, connect, innerconnect, s2s
from .util import subplots

if TYPE_CHECKING:
    from .frequency import Frequency


class Circuit:
    """
    Creates a circuit made of a set of N-ports networks.

    For instructions on how to create Circuit see  :func:`__init__`.

    A Circuit object is representation a circuit assembly of an arbitrary
    number of N-ports networks connected together via an arbitrary topology.

    The algorithm used to calculate the resultant network can be found in [#]_.

    References
    ----------
    .. [#] P. Hallbjörner, Microw. Opt. Technol. Lett. 38, 99 (2003).

    """
    @staticmethod
    def _get_nx():
        """Returns networkx module if available.

        Raises:
        -------
            ImportError: If networkx module is not installed

        Returns:
        --------
            networkx module
        """
        try:
            import networkx as nx
            return nx
        except ImportError as err:
            raise ImportError('networkx package as not been installed and is required.') from err

    class _REDUCE_OPTIONS(TypedDict):
        check_duplication: NotRequired[bool]
        split_ground: NotRequired[bool]
        split_multi: NotRequired[bool]
        max_nports: NotRequired[int]
        dynamic_networks: NotRequired[Sequence[Network]]

    CACHEDPROPERTIES = ('s', 'X', 'X_F', 'C', 'C_F', 'T')

    def __init__(self,
                 connections: list[list[tuple[Network, int]]],
                 name: str | None = None,
                 *,
                 auto_reduce: bool = False, **kwargs: Unpack[_REDUCE_OPTIONS]) -> None:
        """
        Circuit constructor. Creates a circuit made of a set of N-ports networks.

        Parameters
        ----------
        connections : list of list of tuples
            Description of circuit connections.
            Each connection is a described by a list of tuple.
            Each tuple contains (network, network_port_nb).
            Port number indexing starts from zero.
            Network's port not explicitly listed will be treated as matched.
            Network's port listed individually will be treated as open.
        name : string, optional
            Name assigned to the circuit (Network). Default is None.
        auto_reduce : bool, optional
            If True, the circuit will be automatically reduced using :func:`reduce_circuit`.
            This will change the circuit connections description, affecting inner current and voltage distributions.
            Suitable for cases where only the S-parameters of the final circuit ports are of interest. Default is False.
            If `check_duplication`, `split_ground` or `max_nports` are provided as kwargs, `auto_reduce` will be
            automatically set to True, as this indicates an intent to use the `reduce_circuit` method.
        **kwargs :
            keyword arguments passed to `reduce_circuit` method.

            `check_duplication` kwarg controls whether to check the connections have duplicate names. Default is True.

            `split_ground` kwarg controls whether to split the global ground connection to independant.
            Default is False.

            `max_nports` kwarg controls the maximum number of ports of a Network that can be reduced in circuit. If a
            Network in the circuit has a number of ports (nports), using the Network.connect() method to reduce the
            circuit's dimensions becomes less efficient compared to directly calculating it with Circuit.s_external.
            This value depends on the performance of the computer and the scale of the circuit. Default is 20.

            `dynamic_networks` kwarg is a sequence of Networks to be skipped during the reduction process.
            Default is an empty tuple.


        Examples
        --------
        Example of connections between two 1-port networks:
        ::
            connections = [
                [(network1, 0), (network2, 0)],
            ]

        Example of a connection between three 1-port networks connected
        to a single node:
        ::
            connections = [
                [(network1, 0), (network2, 0), (network3, 0)]
            ]

        Example of a connection between two 1-port networks (port1 and port2)
        and two 2-ports networks (ntw1 and ntw2):
        ::
            connections = [
                [(port1, 0), (ntw1, 0)],
                [(ntw1, 1), (ntw2, 0)],
                [(ntw2, 1), (port2, 0)]
            ]

        Example of a connection between three 1-port networks (port1, port2 and port3)
        and a 3-ports network (ntw):
        ::
            connections = [
                [(port1, 0), (ntw, 0)],
                [(port2, 0), (ntw, 1)],
                [(port3, 0), (ntw, 2)]
            ]

        Example of a connection between three 1-port networks (port1, port2 and open)
        and a 3-ports network (ntw):
        ::
            connections = [
                [(port1, 0), (ntw, 0)],
                [(open, 0), (ntw, 1)],
                [(port2, 0), (ntw, 2)]
            ]
            # or equivalently
            connections = [
                [(port1, 0), (ntw, 0)],
                [(ntw, 1)],
                [(port2, 0), (ntw, 2)]
            ]

        Example of a connection between three 1-port networks (port1, port2 and match)
        and a 3-ports network (ntw):
        ::
            connections = [
                [(port1, 0), (ntw, 0)],
                [(match, 0), (ntw, 1)],
                [(port2, 0), (ntw, 2)]
            ]
            # or equivalently
            connections = [
                [(port1, 0), (ntw, 0)],
                [(port2, 0), (ntw, 2)]
            ]

        NB1: Creating 1-port network to be used as a port should be made with :func:`Port`

        NB2: The external ports indexing is defined by the order of appearance of
        the ports in the connections list. Thus, the first network identified
        as a port will be the 1st port of the resulting network (index 0),
        the second network identified as a port will be the second port (index 1),
        etc.

        NB3: When a port of a network is listed individually in the connections, the circuit
        will treat this port with a reflection coefficient of 1, equivalent to an Open.
        Conversely, if a port is not explicitly included in the connections,the circuit will
        treat it as matched.

        """
        self._connections = connections
        self.name = name

        # check if all networks have a name
        for cnx in self.connections:
            for (ntw, _) in cnx:
                if not self._is_named(ntw):
                    raise AttributeError('All Networks must have a name. Faulty network:', ntw)

        # list of networks for initial checks
        ntws = self.networks_list()

        # check if all networks have same frequency
        ref_freq = ntws[0].frequency
        for ntw in ntws:
            if ntw.frequency != ref_freq:
                raise AttributeError('All Networks must have same frequencies')
        # All frequencies are the same, Circuit frequency can be any of the ntw
        self.frequency = ntws[0].frequency

        # Check that a (ntwk, port) combination appears only once in the connexion map
        Circuit.check_duplicate_names(self.connections_list)

        # Get the keyword arguments for the reduce_circuit method
        kwargs_reduce = {k: kwargs[k] for k in self._REDUCE_OPTIONS.__annotations__.keys() if k in kwargs}

        # Reduce the circuit if directly requested or any relevant kwargs are provided
        if auto_reduce or any(kwargs_reduce):
            self.connections = reduce_circuit(self.connections, **kwargs_reduce)

    @property
    def connections(self) -> list[list[tuple[Network, int]]]:
        """
        Circuit connections.

        Description of circuit connections. Each connection is a described by a
        list of tuple. Each tuple contains (network, network_port_nb). Port number
        indexing starts from zero.

        Returns
        -------
        connections : :class:`list[list[tuple[Network, int]]]`
            The circuit connections.

        """
        return self._connections

    @connections.setter
    def connections(self, connections: list[list[tuple[Network, int]]]) -> None:
        """
        Set the circuit connections.

        Parameters
        ----------
        connections : :class:`list[list[tuple[Network, int]]]`
            The circuit connections.

        """
        self._connections = connections

        # Invalidate cached properties
        for item in self.CACHEDPROPERTIES:
            self.__dict__.pop(item, None)

    def update_networks(
        self, networks: tuple[Network],
        name: str | None = None,
        *,
        inplace: bool = False,
        auto_reduce: bool = False, **kwargs: Unpack[_REDUCE_OPTIONS]) -> Circuit | None:
        """
        Update the circuit connections with a new set of networks.

        Parameters
        ----------
        networks : tuple[Network]
            A tuple of Networks to be updated in the circuit.
        name : string, optional
            Name assigned to the circuit (Network). Default is None.
        inplace : bool, optional
            If True, the circuit connections will be updated inplace. Default is False.
        auto_reduce : bool, optional
            If True, the circuit will be automatically reduced using :func:`reduce_circuit`.
            This will change the circuit connections description, affecting inner current and voltage distributions.
            Suitable for cases where only the S-parameters of the final circuit ports are of interest. Default is False.
            If `check_duplication`, `split_ground` or `max_nports` are provided as kwargs, `auto_reduce` will be
            automatically set to True, as this indicates an intent to use the `reduce_circuit` method.
        **kwargs :
            keyword arguments passed to `reduce_circuit` method.

            `check_duplication` kwarg controls whether to check the connections have duplicate names. Default is True.

            `split_ground` kwarg controls whether to split the global ground connection to independant.
            Default is False.

            `max_nports` kwarg controls the maximum number of ports of a Network that can be reduced in circuit. If a
            Network in the circuit has a number of ports (nports), using the Network.connect() method to reduce the
            circuit's dimensions becomes less efficient compared to directly calculating it with Circuit.s_external.
            This value depends on the performance of the computer and the scale of the circuit. Default is 20.

            `dynamic_networks` kwarg is a sequence of Networks to be skipped during the reduction process.
            Default is an empty tuple.

        Returns
        -------
        Circuit or None if inplace=True
            The updated `Circuit` with the specified networks.


        See Also
        --------
        Circuit.__init__ : Circuit construtor method.


        Examples
        --------
        >>> import skrf as rf
        >>> import numpy as np
        >>>
        >>> # Create a series of networks and build a circuit
        >>> connections = [
        ...     ...
        ...     [(ntwk1, 0), (ntwk2, 0)],
        ...     [(ntwk2, 1), (ntwk3, 0)],
        ...     ...
        ... ]
        >>> circuit = rf.Circuit(connections, dynamic_networks=[ntwk2])
        >>>
        >>> # Update the networks' S-parameters
        >>> ntwk2.s = ntwk2.s @ ntwk2.s
        >>>
        >>> # Update the circuit in traditional way (ntwk2 has been updated)
        >>> connections_updated = [
        ...     ...
        ...     [(ntwk1, 0), (ntwk2, 0)],
        ...     [(ntwk2, 1), (ntwk3, 0)],
        ...     ...
        ... ]
        >>>
        >>> circuit_updated_a = rf.Circuit(connections_updated)
        >>>
        >>> # Update the circuit by dynamic networks
        >>> circuit_updated_b = circuit.update_networks(networks=[ntwk2])
        >>>
        >>> np.allclose(circuit_updated_b.s_external, circuit_updated_b.s_external)
        True
        """
        # Get current connection_dict
        cnx_dict = self.networks_dict()

        # Update the connection dict with the new networks
        for ntw in networks:
            if ntw.name in cnx_dict:
                cnx_dict[ntw.name] = ntw
            else:
                raise ValueError(f"Network {ntw.name} not found in circuit.")

        # Update the circuit connections with the new networks
        connections = [
            [(cnx_dict[n.name], p) for n, p in cnx] for cnx in self.connections
        ]

        # Get the keyword arguments for the reduce_circuit method
        kwargs_reduce = {k: kwargs[k] for k in self._REDUCE_OPTIONS.__annotations__.keys() if k in kwargs}

        # Reduce the circuit if directly requested or any relevant kwargs are provided
        if auto_reduce or any(kwargs_reduce):
            connections = reduce_circuit(connections, **kwargs_reduce)

        if inplace:
            self.connections = connections
            return None

        return Circuit(connections=connections, name=name)

    @classmethod
    def check_duplicate_names(cls, connections_list: list[tuple[int, tuple[Network, int]]]):
        """
        Check that a (ntwk, port) combination appears only once in the connexion map
        """
        nodes = [(ntwk.name, port) for (_, (ntwk, port)) in connections_list]
        if len(nodes) > len(set(nodes)):
            duplicate_nodes = [node for node in nodes if nodes.count(node) > 1]
            raise AttributeError(f'Nodes {duplicate_nodes} appears twice in the connection description.')


    def _is_named(self, ntw: Network):
        """
        Return True is the network has a name, False otherwise
        """
        if not ntw.name or ntw.name == '':
            return False
        else:
            return True

    @classmethod
    def _is_port(cls, ntw: Network):
        """
        Return True is the network is a port, False otherwise
        """
        return ntw._ext_attrs.get("_is_circuit_port", False)

    @classmethod
    def _is_ground(cls, ntw: Network):
        """
        Return True is the network is a ground, False otherwise
        """
        return ntw._ext_attrs.get("_is_circuit_ground", False)

    @classmethod
    def _is_open(cls, ntw: Network):
        """
        Return True is the network is a open, False otherwise
        """
        return ntw._ext_attrs.get("_is_circuit_open", False)

    @classmethod
    def Port(cls, frequency: Frequency, name: str, z0: float = 50) -> Network:
        """
        Return a 1-port Network to be used as a Circuit port.

        Parameters
        ----------
        frequency : :class:`~skrf.frequency.Frequency`
            Frequency common to all other networks in the circuit
        name : string
            Name of the port.
        z0 : real, optional
            Characteristic impedance of the port. Default is 50 Ohm.

        Returns
        -------
        port : :class:`~skrf.network.Network` object
            1-port network

        Examples
        --------
        .. ipython::

            @suppress
            In [16]: import skrf as rf

            In [17]: freq = rf.Frequency(start=1, stop=2, npoints=101)

            In [18]: port1 = rf.Circuit.Port(freq, name='Port1')
        """
        _media = media.DefinedGammaZ0(frequency, z0=z0)
        port = _media.match(name=name)
        port._ext_attrs['_is_circuit_port'] = True
        return port

    @classmethod
    def SeriesImpedance(cls, frequency: Frequency, Z: NumberLike, name: str, z0: float = 50) -> Network:
        """
        Return a 2-port network of a series impedance.

        Passing the frequency and name is mandatory.

        Parameters
        ----------
        frequency : :class:`~skrf.frequency.Frequency`
            Frequency common to all other networks in the circuit
        Z : complex array of shape n_freqs or complex
            Impedance
        name : string
            Name of the series impedance
        z0 : real, optional
            Characteristic impedance of the port. Default is 50 Ohm.

        Returns
        -------
        serie_impedance : :class:`~skrf.network.Network` object
            2-port network

        Examples
        --------
        .. ipython::

            @suppress
            In [16]: import skrf as rf

            In [17]: freq = rf.Frequency(start=1, stop=2, npoints=101)

            In [18]: open = rf.Circuit.SeriesImpedance(freq, rf.INF, name='series_impedance')

        """
        A = np.zeros(shape=(len(frequency), 2, 2), dtype=complex)
        A[:, 0, 0] = 1
        A[:, 0, 1] = Z
        A[:, 1, 0] = 0
        A[:, 1, 1] = 1
        ntw = Network(a=A, frequency=frequency, z0=z0, name=name)
        return ntw

    @classmethod
    def ShuntAdmittance(cls, frequency: Frequency, Y: NumberLike, name: str, z0: float = 50) -> Network:
        """
        Return a 2-port network of a shunt admittance.

        Passing the frequency and name is mandatory.

        Parameters
        ----------
        frequency : :class:`~skrf.frequency.Frequency`
            Frequency common to all other networks in the circuit
        Y : complex array of shape n_freqs or complex
            Admittance
        name : string
            Name of the shunt admittance
        z0 : real, optional
            Characteristic impedance of the port. Default is 50 Ohm.

        Returns
        -------
        shunt_admittance : :class:`~skrf.network.Network` object
            2-port network

        Examples
        --------
        .. ipython::

            @suppress
            In [16]: import skrf as rf

            In [17]: freq = rf.Frequency(start=1, stop=2, npoints=101)

            In [18]: short = rf.Circuit.ShuntAdmittance(freq, rf.INF, name='shunt_admittance')

        """
        A = np.zeros(shape=(len(frequency), 2, 2), dtype=complex)
        A[:, 0, 0] = 1
        A[:, 0, 1] = 0
        A[:, 1, 0] = Y
        A[:, 1, 1] = 1
        ntw = Network(a=A, frequency=frequency, z0=z0, name=name)
        return ntw

    @classmethod
    def Ground(cls, frequency: Frequency, name: str, z0: float = 50) -> Network:
        """
        Return a 1-port network of a grounded link.

        Passing the frequency and a name is mandatory.

        The ground link is implemented by media.short object.

        Parameters
        ----------
        frequency : :class:`~skrf.frequency.Frequency`
            Frequency common to all other networks in the circuit
        name : string
            Name of the ground.
        z0 : real, optional
            Characteristic impedance of the port. Default is 50 Ohm.

        Returns
        -------
        ground : :class:`~skrf.network.Network` object
            1-port network

        Examples
        --------
        .. ipython::

            @suppress
            In [16]: import skrf as rf

            In [17]: freq = rf.Frequency(start=1, stop=2, npoints=101)

            In [18]: ground = rf.Circuit.Ground(freq, name='GND')

        """
        _media = media.DefinedGammaZ0(frequency, z0=z0)
        ground = _media.short(name=name)
        ground._ext_attrs['_is_circuit_ground'] = True
        return ground

    @classmethod
    def Open(cls, frequency: Frequency, name: str, z0: float = 50) -> Network:
        """
        Return a 1-port network of an open link.

        Passing the frequency and name is mandatory.

        The open link is implemented by media.open object.

        Parameters
        ----------
        frequency : :class:`~skrf.frequency.Frequency`
            Frequency common to all other networks in the circuit
        name : string
            Name of the open.
        z0 : real, optional
            Characteristic impedance of the port. Default is 50 Ohm.

        Returns
        -------
        open : :class:`~skrf.network.Network` object
            1-port network

        Examples
        --------
        .. ipython::

            @suppress
            In [16]: import skrf as rf

            In [17]: freq = rf.Frequency(start=1, stop=2, npoints=101)

            In [18]: open = rf.Circuit.Open(freq, name='open')

        """
        _media = media.DefinedGammaZ0(frequency, z0=z0)
        Open = _media.open(name=name)
        Open._ext_attrs['_is_circuit_open'] = True
        return Open

    def networks_dict(self,
                      connections: list[list[tuple[Network, int]]] | None = None,
                      min_nports: int = 1) -> dict[str, Network]:
        """
        Return the dictionary of Networks from the connection setup X.

        Parameters
        ----------
        connections : List, optional
            connections list, by default None (then uses the `self.connections`)
        min_nports : int, optional
            min number of ports, by default 1

        Returns
        -------
        dict
            Dictionnary of Networks
        """
        if not connections:
            connections = self.connections

        ntws: list[Network] = []
        for cnx in connections:
            for (ntw, _port) in cnx:
                ntws.append(ntw)
        return {ntw.name: ntw for ntw in ntws  if ntw.nports >= min_nports}

    def networks_list(self,
                      connections: list[list[tuple[Network, int]]] | None = None,
                      min_nports: int = 1) -> list[Network]:
        """
        Return a list of unique networks (sorted by appearing order in connections).

        Parameters
        ----------
        connections : List, optional
            connections list, by default None (then uses the `self.connections`)
        min_nports : int, optional
            min number of ports, by default 1

        Returns
        -------
        list
            List of unique networks
        """
        if not connections:
            connections = self.connections

        ntw_dict = self.networks_dict(connections)
        return [ntw for ntw in ntw_dict.values() if ntw.nports >= min_nports]

    @property
    def connections_nb(self) -> int:
        """
        Return the number of intersections in the circuit.
        """
        return len(self.connections)

    @property
    def connections_list(self) -> list[tuple[int, tuple[Network, int]]]:
        """
        Return the full list of connections, including intersections.

        The resulting list if of the form:
        ::
            [
             [connexion_number, connexion],
             [connexion_number, connexion],
             ...
            ]
        """
        return list(enumerate(chain.from_iterable(self.connections)))

    @property
    def networks_nb(self) -> int:
        """
        Return the number of connected networks (port excluded).
        """
        return len(self.networks_list(self.connections))

    @property
    def nodes_nb(self) -> int:
        """
        Return the number of nodes in the circuit.
        """
        return self.connections_nb + self.networks_nb

    @property
    def dim(self) -> int:
        """
        Return the dimension of the C, X and global S matrices.

        It correspond to the sum of all connections.
        """
        return np.sum([len(cnx) for cnx in self.connections])

    @property
    def G(self):
        """
        Generate the graph of the circuit. Convenience shortname for :func:`graph`.
        """
        return self.graph()

    def graph(self):
        """
        Generate the graph of the circuit.

        Returns
        -------
        G: :class:`networkx.Graph`
            graph object [#]_ .

        References
        ----------
        .. [#] https://networkx.github.io/
        """

        nx = self._get_nx()
        G = nx.Graph()
        # Adding network nodes
        G.add_nodes_from([it for it in self.networks_dict(self.connections)])

        # Adding edges in the graph between connections and networks
        for (idx, cnx) in enumerate(self.connections):
            cnx_name = 'X'+str(idx)
            # Adding connection nodes and edges
            G.add_node(cnx_name)
            for (ntw, _ntw_port) in cnx:
                ntw_name = ntw.name
                G.add_edge(cnx_name, ntw_name)
        return G

    def is_connected(self) -> bool:
        """
        Check if the circuit's graph is connected.

        Check if every pair of vertices in the graph is connected.
        """

        nx = self._get_nx()
        return nx.algorithms.components.is_connected(self.G)


    @property
    def intersections_dict(self) -> dict:
        """
        Return a dictionary of all intersections with associated ports and z0:

        ::
            { k: [(ntw1_name, ntw1_port), (ntw1_z0, ntw2_name, ntw2_port), ntw2_z0], ... }
        """
        inter_dict = {}
        # for k in range(self.connections_nb):
        #     # get all edges connected to intersection Xk
        #     inter_dict[k] = list(nx.algorithms.boundary.edge_boundary(self.G, ('X'+str(k),) ))

        for (k, cnx) in enumerate(self.connections):
            inter_dict[k] = [(ntw, ntw_port, ntw.z0[0, ntw_port]) \
                                  for (ntw, ntw_port) in cnx]
        return inter_dict

    @property
    def edges(self) -> list:
        """
        Return the list of all circuit connections
        """
        return list(self.G.edges)

    @property
    def edge_labels(self) -> dict:
        """
        Return a dictionary describing the port and z0 of all graph edges.

        Dictionary is in the form:
        ::
            {('ntw1_name', 'X0'): '3 (50+0j)',
             ('ntw2_name', 'X0'): '0 (50+0j)',
             ('ntw2_name', 'X1'): '2 (50+0j)', ... }

        which can be used in `networkx.draw_networkx_edge_labels`
        """
        # for all intersections,
        # get the N interconnected networks and associated ports and z0
        # and forge the edge label dictionary containing labels between
        # two nodes
        edge_labels = {}
        for it in self.intersections_dict.items():
            k, cnx = it
            for idx in range(len(cnx)):
                ntw, ntw_port, ntw_z0 = cnx[idx]
                #ntw_z0 = ntw.z0[0,ntw_port]
                edge_labels[(ntw.name, 'X'+str(k))] = str(ntw_port)+'\n'+\
                                        str(np.round(ntw_z0, decimals=1))

        return edge_labels


    def _Xk(self, cnx_k: list[tuple[Network, int]], order: MemoryLayoutT = 'C') -> np.ndarray:
        """
        Return the scattering matrices [X]_k of the individual intersections k.
        The results in [#]_ do not agree due to an error in the formula (3)
        for mismatched intersections.

        Parameters
        ----------
        cnx_k : list of tuples
            each tuple contains (network, port)
        order: str, optional
            'C' or 'F' for row-major or column-major order of the output array.
            Default is 'C'.

        Returns
        -------
        Xs : :class:`numpy.ndarray`
            shape `f x n x n`

        References
        ----------
        .. [#] P. Hallbjörner, Microw. Opt. Technol. Lett. 38, 99 (2003).
        """

        y0s = np.array([1/ntw.z0[:,ntw_port] for (ntw, ntw_port) in cnx_k]).T
        y_k = y0s.sum(axis=1)

        Xs = np.zeros((len(self.frequency), len(cnx_k), len(cnx_k)), dtype='complex', order=order)

        Xs = 2 *np.sqrt(np.einsum('ij,ik->ijk', y0s, y0s)) / y_k[:, None, None]
        np.einsum('kii->ki', Xs)[:] -= 1  # Sii
        return Xs

    def _X(self, order: MemoryLayoutT = 'C') -> np.ndarray:
        """
        Return the concatenated intersection matrix [X] of the circuit.

        It is composed of the individual intersection matrices [X]_k assembled
        by block diagonal.

        Parameters
        ----------
            order : str, optional
                'C' or 'F' for row-major or column-major order of the output array.
                Default is 'C'.

        Returns
        -------
        X : :class:`numpy.ndarray`

        Note
        ----
        The block diagonal matrix [X] has a numerical bottleneck that depends on the
        order specified:
        - When 'order' is 'C', the creation of the block diagonal matrix [X] in row-major
          order is a bottleneck. This is because the creation process is port-by-port,
          which does not align well with the row-major memory layout, leading to
          suboptimal performance.
        - When 'order' is 'F', the computation of the block diagonal matrix [X] in column-major
          order is a bottleneck. Numpy generally optimizes operators for 'C' order, which
          lead to performance issues when using 'F' order.
        """
        Xks = [self._Xk(cnx, order) for cnx in self.connections]

        Xf = np.zeros((len(self.frequency), self.dim, self.dim), dtype='complex', order=order)
        off = np.array([0, 0])
        for Xk in Xks:
            Xf[:, off[0]:off[0] + Xk.shape[1], off[1]:off[1]+Xk.shape[2]] = Xk
            off += Xk.shape[1:]

        return Xf

    @cached_property
    def X(self) -> np.ndarray:
        """
        Return the concatenated intersection matrix [X] of the circuit in C-order.

        It is composed of the individual intersection matrices [X]_k assembled
        by block diagonal.

        Returns
        -------
        X : :class:`numpy.ndarray`

        Note
        ----
        There is a numerical bottleneck in this function,
        when creating the block diagonal matrice [X] from the [X]_k matrices.
        """
        # Check if X_F is alread computed, if so, convert it to C-order
        if self.__dict__.get('X_F', None) is not None:
            return np.ascontiguousarray(self.X_F)

        return self._X()

    @cached_property
    def X_F(self) -> np.ndarray:
        """
        Return the concatenated intersection matrix [X] of the circuit in F-order.
        It is composed of the individual intersection matrices [X]_k assembled
        by block diagonal. The results of this function are the same as :func:`X`
        but in F-order.

        Returns
        -------
        X : :class:`numpy.ndarray`

        Note
        ----
        F-order has a numerical bottleneck in matrix operations, but the assignment
        is more efficient.
        """
        # Check if X is alread computed, if so, convert it to F-order
        if self.__dict__.get('X', None) is not None:
            return np.asfortranarray(self.X)

        return self._X('F')

    @cached_property
    def C(self) -> np.ndarray:
        """
        Return the global scattering matrix [C] of the networks in C-order.

        Returns
        -------
        S : :class:`numpy.ndarray`
            Global scattering matrix of the networks.
            Shape `f x (nb_inter*nb_n) x (nb_inter*nb_n)`
        """
        # Check if C_F is alread computed, if so, convert it to C-order
        if self.__dict__.get('C_F', None) is not None:
            return np.ascontiguousarray(self.C_F)

        return self._C()

    @cached_property
    def C_F(self) -> np.ndarray:
        """
        Return the global scattering matrix [C] of the networks in F-order. The results
        of this function are the same as :func:`C` but in F-order.

        Returns
        -------
        S : :class:`numpy.ndarray`
            Global scattering matrix of the networks.
            Shape `f x (nb_inter*nb_n) x (nb_inter*nb_n)`

        Note
        ----
        F-order has a numerical bottleneck in matrix operations, but the assignment
        is more efficient.
        """
        # Check if C is alread computed, if so, convert it to F-order
        if self.__dict__.get('C', None) is not None:
            return np.asfortranarray(self.C)

        return self._C('F')

    def _C(self, order: MemoryLayoutT = 'C') -> np.ndarray:
        """
        Return the global scattering matrix [C] of the networks.

        Args:
            order : str, optional
                'C' or 'F' for row-major or column-major order of the output array.
                Default is 'C'.

        Returns
        -------
        S : :class:`numpy.ndarray`
            Global scattering matrix of the networks.
            Shape `f x (nb_inter*nb_n) x (nb_inter*nb_n)`
        """
        # list all networks which are not considered as "ports",
        ntws = {k:v for k,v in self.networks_dict().items() if not Circuit._is_port(v)}

        # generate the port reordering indexes from each connections
        ntws_ports_reordering = {ntw:[] for ntw in ntws}
        for (idx_cnx, (ntw, ntw_port)) in self.connections_list:
            if ntw.name in ntws.keys():
                ntws_ports_reordering[ntw.name].append([ntw_port, idx_cnx])

        # re-ordering scattering parameters
        S = np.zeros((len(self.frequency), self.dim, self.dim), dtype='complex', order=order)

        for (ntw_name, ntw_ports) in ntws_ports_reordering.items():
            # get the port re-ordering indexes (from -> to)
            ntw_ports = np.array(ntw_ports)

            # port permutations
            from_port = ntw_ports[:,0]
            to_port = ntw_ports[:,1]

            for (_from, _to) in zip(from_port, to_port):
                S[:, _to, to_port] = ntws[ntw_name].s_traveling[:, _from, from_port]

        return S  # shape (nb_frequency, nb_inter*nb_n, nb_inter*nb_n)

    @cached_property
    def T(self) -> np.ndarray:
        """
        Return the matrix of multiplication of the global scattering matrix [C] and concatenated
        intersection matrix [X] of the networks, that is [T] = - [C] @ [X].

        Returns
        -------
        T : :class:`numpy.ndarray`
            Multiplication of the global scattering matrix [C] and concatenated intersection matrix
            [X] of the networks. In practice, F-contiguous [C_F] and [X_F] are used.
            Shape `f x (nb_inter*nb_n) x (nb_inter*nb_n)`

        Note
        ----
        This is an auxiliary matrix used to break the numerical bottleneck of [C_F] @ [X_F] using the
        mathematical feature of block diagonal matrice [X].
        """
        X, C = self.X_F, self.C_F

        # T will be fully updated, so `empty_like` is safe and faster
        T = np.empty_like(X, dtype="complex", order='F')

        # Precompute the sizes of connections and slices for each intersection
        cnx_size = [len(cnx) for cnx in self.connections]
        slice_ = np.cumsum([0] + cnx_size)
        slices = [slice(slice_[i], slice_[i+1]) for i in range(len(cnx_size))]

        # Perform the multiplication
        for j_slice in slices:
            # Get the Block diagonal part of X and corresponding C matrix buffer
            X_jj = - X[:, j_slice, j_slice]
            C_j = C[:, :, j_slice]

            # Perform the multiplication
            for i_slice in slices:
                T[:, i_slice, j_slice] = C_j[:, i_slice, :] @ X_jj

        return T

    @cached_property
    def s(self) -> np.ndarray:
        """
        Return the global scattering parameters of the circuit.

        Return the scattering parameters of both "inner" and "outer" ports.

        Returns
        -------
        S : :class:`numpy.ndarray`
            global scattering parameters of the circuit.
        """
        X = self.X
        return X @ np.linalg.inv(np.identity(self.dim) - self.C @ X)

    @property
    def port_indexes(self) -> list[int]:
        """
        Return the indexes of the "external" ports.

        Returns
        -------
        port_indexes : list
        """
        port_indexes = []
        for (idx_cnx, (ntw, _)) in enumerate(chain.from_iterable(self.connections)):
            if Circuit._is_port(ntw):
                port_indexes.append(idx_cnx)
        return port_indexes

    def _cnx_z0(self, cnx_k: list[tuple]) -> np.ndarray:
        """
        Return the characteristic impedances of a specific connections.

        Parameters
        ----------
        cnx_k : list of tuples
            each tuple contains (network, port)

        Returns
        -------
        z0s : :class:`numpy.ndarray`
            shape `f x nb_ports_at_cnx`
        """
        z0s = []
        for (ntw, ntw_port) in cnx_k:
            z0s.append(ntw.z0[:,ntw_port])

        return np.array(z0s).T  # shape (nb_freq, nb_ports_at_cnx)

    @property
    def port_z0(self) -> np.ndarray:
        """
        Return the external port impedances.

        Returns
        -------
        z0s : :class:`numpy.ndarray`
            shape `f x nb_ports`
        """
        z0s = []
        for cnx in self.connections:
            for (ntw, ntw_port) in cnx:
                z0s.append(ntw.z0[:,ntw_port])

        return np.array(z0s)[self.port_indexes, :].T  # shape (nb_freq, nb_ports)

    @property
    def s_external(self) -> np.ndarray:
        """
        Return the scattering parameters for the external ports.

        Returns
        -------
        S : :class:`numpy.ndarray`
            Scattering parameters of the circuit for the external ports.
            Shape `f x nb_ports x nb_ports`
        """
        # The external S-matrix is the submatrix corresponding to external ports:
        # port_indexes = self.port_indexes
        # a, b = np.meshgrid(port_indexes, port_indexes, indexing='ij')
        # S_ext = self.s[:, a, b]

        # Instead of calculating all S-parameters and taking a submatrix,
        # the following faster approach only calculates external the S-parameters
        # from block-matrix operations.
        # generate index lists of internal and external ports
        port_indexes = self.port_indexes
        in_idxs = [(i,) for i in range(self.dim) if i not in port_indexes]
        ext_idxs = [(i,) for i in port_indexes]
        ext_l, in_l = len(ext_idxs), len(in_idxs)

        # generate index slices for each sub-matrices
        idx_a, idx_b, idx_c, idx_d = (
            np.repeat(i, l, axis=1)
            for i, l in (
                (ext_idxs, ext_l),
                (ext_idxs, in_l),
                (in_idxs, ext_l),
                (in_idxs, in_l),
            )
        )

        # sub-matrices index, Matrix = [[A, B], [C, D]]]
        A_idx = (slice(None), idx_a, idx_a.T)
        B_idx = (slice(None), idx_b, idx_c.T)
        C_idx = (slice(None), idx_c, idx_b.T)
        D_idx = (slice(None), idx_d, idx_d.T)

        # Get the buffer of global matrix in f-order [X_T] and intermediate temporary matrix [T]
        # [T] = - [C] @ [X]
        x, t = self.X_F, np.array(self.T)
        np.einsum('...ii->...i', t)[:] += 1

        # Get the sub-matrices of inverse of intermediate temporary matrix t
        # The method np.linalg.solve(A, B) is equivalent to np.inv(A) @ B, but more efficient
        tmp_mat = np.linalg.solve(t[D_idx], t[C_idx])

        # Get the external S-parameters for the external ports
        # Calculated by multiplying the sub-matrices of x and t
        S_ext = (x[A_idx] - x[B_idx] @ tmp_mat) @ np.linalg.inv(
            t[A_idx] - t[B_idx] @ tmp_mat
        )

        S_ext = s2s(S_ext, self.port_z0, S_DEF_DEFAULT, 'traveling')
        return S_ext  # shape (nb_frequency, nb_ports, nb_ports)

    @property
    def network(self) -> Network:
        """
        Return the Network associated to external ports.

        Returns
        -------
        ntw : :class:`~skrf.network.Network`
            Network associated to external ports
        """
        return Network(frequency = self.frequency, z0 = self.port_z0,
                      s = self.s_external, name = self.name)

    def s_active(self, a: NumberLike) -> np.ndarray:
        r"""
        Return "active" s-parameters of the circuit's network for a defined wave excitation `a`.

        The "active" s-parameter at a port is the reflection coefficients
        when other ports are excited. It is an important quantity for active
        phased array antennas.

        Active s-parameters are defined by [#]_:

        .. math::

            \mathrm{active}(s)_{mn} = \sum_i s_{mi} \frac{a_i}{a_n}

        Parameters
        ----------
        a : complex array of shape (n_ports)
            forward wave complex amplitude (power-wave formulation [#]_)

        Returns
        -------
        s_act : complex array of shape (n_freqs, n_ports)
            active s-parameters for the excitation a


        References
        ----------
        .. [#] D. M. Pozar, IEEE Trans. Antennas Propag. 42, 1176 (1994).

        .. [#] D. Williams, IEEE Microw. Mag. 14, 38 (2013).

        """
        return self.network.s_active(a)

    def z_active(self, a: NumberLike) -> np.ndarray:
        r"""
        Return the "active" Z-parameters of the circuit's network for a defined wave excitation a.

        The "active" Z-parameters are defined by:

        .. math::

            \mathrm{active}(z)_{m} = z_{0,m} \frac{1 + \mathrm{active}(s)_m}{1 - \mathrm{active}(s)_m}

        where :math:`z_{0,m}` is the characteristic impedance and
        :math:`\mathrm{active}(s)_m` the active S-parameter of port :math:`m`.

        Parameters
        ----------
        a : complex array of shape (n_ports)
            forward wave complex amplitude

        Returns
        -------
        z_act : complex array of shape (nfreqs, nports)
            active Z-parameters for the excitation a

        See Also
        --------
            s_active : active S-parameters
            y_active : active Y-parameters
            vswr_active : active VSWR
        """
        return self.network.z_active(a)

    def y_active(self, a: NumberLike) -> np.ndarray:
        r"""
        Return the "active" Y-parameters of the circuit's network for a defined wave excitation a.

        The "active" Y-parameters are defined by:

        .. math::

            \mathrm{active}(y)_{m} = y_{0,m} \frac{1 - \mathrm{active}(s)_m}{1 + \mathrm{active}(s)_m}

        where :math:`y_{0,m}` is the characteristic admittance and
        :math:`\mathrm{active}(s)_m` the active S-parameter of port :math:`m`.

        Parameters
        ----------
        a : complex array of shape (n_ports)
            forward wave complex amplitude

        Returns
        -------
        y_act : complex array of shape (nfreqs, nports)
            active Y-parameters for the excitation a

        See Also
        --------
            s_active : active S-parameters
            z_active : active Z-parameters
            vswr_active : active VSWR
        """
        return self.network.y_active(a)

    def vswr_active(self, a: NumberLike) -> np.ndarray:
        r"""
        Return the "active" VSWR of the circuit's network for a defined wave excitation a.

        The "active" VSWR is defined by :

        .. math::

            \mathrm{active}(vswr)_{m} = \frac{1 + |\mathrm{active}(s)_m|}{1 - |\mathrm{active}(s)_m|}

        where :math:`\mathrm{active}(s)_m` the active S-parameter of port :math:`m`.

        Parameters
        ----------
        a : complex array of shape (n_ports)
            forward wave complex amplitude

        Returns
        -------
        vswr_act : complex array of shape (nfreqs, nports)
            active VSWR for the excitation a

        See Also
        --------
            s_active : active S-parameters
            z_active : active Z-parameters
            y_active : active Y-parameters
        """
        return self.network.vswr_active(a)

    @property
    def z0(self) -> np.ndarray:
        """
        Characteristic impedances of "internal" ports.

        Returns
        -------
        z0 : complex array of shape (nfreqs, nports)
            Characteristic impedances of both "inner" and "outer" ports

        """
        z0s = []
        for _cnx_idx, (ntw, ntw_port) in self.connections_list:
            z0s.append(ntw.z0[:,ntw_port])
        return np.array(z0s).T

    @property
    def connections_pair(self) -> list:
        """
        List the connections by pair.

        Each connection in the circuit is between a specific pair of two
        (networks, port, z0).

        Returns
        -------
        connections_pair : list
            list of pair of connections

        """
        return [self.connections_list[i:i+2] for i in range(0, len(self.connections_list), 2)]


    @property
    def _currents_directions(self) -> np.ndarray:
        """
        Create a array of indices to define the sign of the current.

        The currents are defined positive when entering an internal network.

        Returns
        -------
        directions : array of int (nports, 2)

        Note
        ----
        This function is used in internal currents and voltages calculations.

        """
        directions = np.zeros((self.dim,2), dtype='int')
        for cnx_pair in self.connections_pair:
            (cnx_idx_A, cnx_A), (cnx_idx_B, cnx_B) = cnx_pair
            directions[cnx_idx_A,:] = cnx_idx_A, cnx_idx_B
            directions[cnx_idx_B,:] = cnx_idx_B, cnx_idx_A
        return directions

    def _a(self, a_external: NumberLike) -> np.ndarray:
        """
        Wave input array at "internal" ports.

        Parameters
        ----------
        a_external : array
            power-wave input vector at ports

        Returns
        -------
        a_internal : array
            Wave input array at internal ports

        """
        # create a zero array and fill the values corresponding to ports
        a_internal = np.zeros(self.dim, dtype='complex')
        a_internal[self.port_indexes] = a_external
        return a_internal

    def _a_external(self, power: NumberLike, phase: NumberLike) -> np.ndarray:
        r"""
        Wave input array at Circuit's ports ("external" ports).

        The array is defined from power and phase by:

        .. math::

            a = \sqrt(2 P_{in} ) e^{j \phi}

        The factor 2 is in order to deal with peak values.

        Parameters
        ----------
        power : list or array
            Input power at external ports in Watts [W]
        phase : list or array
            Input phase at external ports in radian [rad]

        NB: the size of the power and phase array should match the number of ports

        Returns
        -------
        a_external: array
            Wave input array  at Circuit's ports

        """
        if len(power) != len(self.port_indexes):
            raise ValueError('Length of power array does not match the number of ports of the circuit.')
        if len(phase) != len(self.port_indexes):
            raise ValueError('Length of phase array does not match the number of ports of the circuit.')
        return np.sqrt(2*np.array(power))*np.exp(1j*np.array(phase))

    def _b(self, a_internal: NumberLike) -> np.ndarray:
        """
        Wave output array at "internal" ports

        Parameters
        ----------
        a_internal : array
            Wave input array at internal ports

        Returns
        -------
        b_internal : array
            Wave output array at internal ports

        Note
        ----
        Wave input array at internal ports can be derived from power and
        phase excitation at "external" ports using `_a(power, phase)` method.

        """
        return self.s @ a_internal

    def currents(self, power: NumberLike, phase: NumberLike) -> np.ndarray:
        """
        Currents at internal ports.

        NB: current direction is defined as positive when entering a node.

        NB: external current sign are opposite than corresponding internal ones,
            as the internal currents are actually flowing into the "port" networks

        Parameters
        ----------
        power : list or array
            Input power at external ports in Watts [W]
        phase : list or array
            Input phase at external ports in radian [rad]

        Returns
        -------
        I : complex array of shape (nfreqs, nports)
            Currents in Amperes [A] (peak) at internal ports.

        """
        a = self._a(self._a_external(power, phase))
        b = self._b(a)
        z0s = self.z0
        i, Is = 0, np.zeros_like(z0s)

        for cnx in self.connections:
            cnx_len = len(cnx)
            z0_segment = z0s[:, i : i + cnx_len]
            tot_shunt_z0 = (1 / z0_segment).sum(axis=1)
            Ij = np.zeros_like(z0_segment)

            # Calculate the ports' output current through the output wave
            for j in range(cnx_len):
                in_z0 = z0_segment[:, j]
                out_z0 = 1 / (tot_shunt_z0 - 1 / in_z0)
                tau = (2 * out_z0) / (out_z0 + in_z0)
                Ij[:, j] = (b[:, i + j] / np.sqrt(in_z0)) * tau

            # The current of each port is different in the same node
            # The ports' current should take into account the output current of each port in the node
            for j in range(cnx_len):
                in_z0 = z0_segment[:, j]
                out_z0 = 1 / (tot_shunt_z0 - 1 / in_z0)
                Itmp = np.zeros_like(Is[:, i + j])
                for k in range(cnx_len):
                    tmp_z0 = z0_segment[:, k]
                    if j == k:
                        Itmp += Ij[:, k] * (tmp_z0 / out_z0)
                    else:
                        Itmp -= Ij[:, k] * (tmp_z0 / in_z0)

                Is[:, i + j] = Itmp

            i += cnx_len

        return Is


    def voltages(self, power: NumberLike, phase: NumberLike) -> np.ndarray:
        """
        Voltages at internal ports.

        Parameters
        ----------
        power : list or array
            Input power at external ports in Watts [W]
        phase : list or array
            Input phase at external ports in radian [rad]

        Returns
        -------
        V : complex array of shape (nfreqs, nports)
            Voltages in Volt [V] (peak) at internal ports.

        """
        a = self._a(self._a_external(power, phase))
        b = self._b(a)
        z0s = self.z0
        i, Vs = 0, np.zeros_like(z0s)

        for cnx in self.connections:
            cnx_len = len(cnx)
            z0_segment = z0s[:, i : i + cnx_len]
            tot_shunt_z0 = (1 / z0_segment).sum(axis=1)
            Vk = np.zeros(shape=z0s.shape[0], dtype="complex128")

            # Node voltage is the summation of each ports' outwave voltage
            # The voltage of each port in the same node is consistent
            for j in range(cnx_len):
                in_z0 = z0_segment[:, j]
                out_z0 = 1 / (tot_shunt_z0 - 1 / in_z0)
                tau = (2 * out_z0) / (out_z0 + in_z0)
                Vk += (b[:, i + j] * np.sqrt(in_z0)) * tau

            Vs[:, i : i + cnx_len] = Vk[:, None]
            i += cnx_len

        return Vs

    def currents_external(self, power: NumberLike, phase: NumberLike) -> np.ndarray:
        """
        Currents at external ports.

        NB: current direction is defined positive when "entering" into port.

        Parameters
        ----------
        power : list or array
            Input power at external ports in Watts [W]
        phase : list or array
            Input phase at external ports in radian [rad]

        Returns
        -------
        I : complex array of shape (nfreqs, nports)
            Currents in Amperes [A] (peak) at external ports.

        """
        a = self._a(self._a_external(power, phase))
        b = self._b(a)
        z0s = self.z0
        Is = []
        for port_idx in self.port_indexes:
            Is.append((a[port_idx] - b[:,port_idx])/np.sqrt(z0s[:,port_idx]))
        return np.array(Is).T

    def voltages_external(self, power: NumberLike, phase: NumberLike) -> np.ndarray:
        """
        Voltages at external ports

        Parameters
        ----------
        power : list or array
            Input power at external ports in Watts [W]
        phase : list or array
            Input phase at external ports in radian [rad]

        Returns
        -------
        V : complex array of shape (nfreqs, nports)
            Voltages in Volt [V] (peak)  at ports

        """
        a = self._a(self._a_external(power, phase))
        b = self._b(a)
        z0s = self.z0
        Vs = []
        for port_idx in self.port_indexes:
            Vs.append((a[port_idx] + b[:,port_idx])*np.sqrt(z0s[:,port_idx]))
        return np.array(Vs).T

    def plot_graph(self, **kwargs):
        """
        Plot the graph of the circuit using networkx drawing capabilities.

        Customisation options with default values:
        ::
            'network_shape': 's'
            'network_color': 'gray'
            'network_size', 300
            'network_fontsize': 7
            'inter_shape': 'o'
            'inter_color': 'lightblue'
            'inter_size', 300
            'port_shape': '>'
            'port_color': 'red'
            'port_size', 300
            'port_fontsize': 7
            'edges_fontsize': 5
            'network_labels': False
            'edge_labels': False
            'inter_labels': False
            'port_labels': False
            'label_shift_x': 0
            'label_shift_y': 0

        """

        nx = self._get_nx()
        G = self.G

        # default values
        network_labels = kwargs.pop('network_labels', False)
        network_shape = kwargs.pop('network_shape', 's')
        network_color = kwargs.pop('network_color', 'gray')
        network_fontsize = kwargs.pop('network_fontsize', 7)
        network_size = kwargs.pop('network_size', 300)
        inter_labels = kwargs.pop('inter_labels', False)
        inter_shape = kwargs.pop('inter_shape', 'o')
        inter_color = kwargs.pop('inter_color', 'lightblue')
        inter_size = kwargs.pop('inter_size', 300)
        port_labels = kwargs.pop('port_labels', False)
        port_shape = kwargs.pop('port_shape', '>')
        port_color = kwargs.pop('port_color', 'red')
        port_size = kwargs.pop('port_size', 300)
        port_fontsize = kwargs.pop('port_fontsize', 7)
        edge_labels = kwargs.pop('edge_labels', False)
        edge_fontsize = kwargs.pop('edge_fontsize', 5)
        label_shift_x = kwargs.pop('label_shift_x', 0)
        label_shift_y = kwargs.pop('label_shift_y', 0)


        # sort between network nodes and port nodes
        all_ntw_names = [ntw.name for ntw in self.networks_list()]
        port_names = [ntw_name for ntw_name in all_ntw_names if 'port' in ntw_name]
        ntw_names = [ntw_name for ntw_name in all_ntw_names if 'port' not in ntw_name]
        # generate connecting nodes names
        int_names = ['X'+str(k) for k in range(self.connections_nb)]

        fig, ax = subplots(figsize=(10,8))

        pos = nx.spring_layout(G)

        # draw Networks
        nx.draw_networkx_nodes(G, pos, port_names, ax=ax,
                            node_size=port_size,
                            node_color=port_color, node_shape=port_shape)
        nx.draw_networkx_nodes(G, pos, ntw_names, ax=ax,
                            node_size=network_size,
                            node_color=network_color, node_shape=network_shape)
        # draw intersections
        nx.draw_networkx_nodes(G, pos, int_names, ax=ax,
                            node_size=inter_size,
                            node_color=inter_color, node_shape=inter_shape)
        # labels shifts
        pos_labels = {}
        for node, coords in pos.items():
            pos_labels[node] = (coords[0] + label_shift_x,
                                coords[1] + label_shift_y)

        # network labels
        if network_labels:
            network_labels = {lab:lab for lab in ntw_names}

            nx.draw_networkx_labels(G, pos_labels, labels=network_labels,
                                    font_size=network_fontsize, ax=ax)

        # intersection labels
        if inter_labels:
            inter_labels = {'X'+str(k):'X'+str(k) for k in range(self.connections_nb)}

            nx.draw_networkx_labels(G, pos_labels, labels=inter_labels,
                                    font_size=network_fontsize, ax=ax)

        # port labels
        if port_labels:
            port_labels = {lab:lab for lab in port_names}

            nx.draw_networkx_labels(G, pos_labels, labels=port_labels,
                                    font_size=port_fontsize, ax=ax)

        # draw edges
        nx.draw_networkx_edges(G, pos, ax=ax)
        if edge_labels:
            edge_labels = self.edge_labels
            nx.draw_networkx_edge_labels(G, pos,
                                        edge_labels=edge_labels, label_pos=0.5,
                                        font_size=edge_fontsize, ax=ax)
        # remove x and y axis and labels
        ax.axis('off')
        fig.tight_layout()

## Functions operating on Circuit
def reduce_circuit(connections: list[list[tuple[Network, int]]],
                   check_duplication: bool = True,
                   split_ground: bool = True,
                   split_multi: bool = False,
                   max_nports: int = 20,
                   dynamic_networks: Sequence[Network] = tuple()) -> list[list[tuple[Network, int]]]:
    """
    Return a reduced equivalent circuit connections with fewer components.

    The reduced equivalent circuit connections allows faster calculation of the circuit network.


    Parameters
    ----------
    connections : list[list[tuple[Network, int]]].
            The connection list to reduce.
    check_duplication : bool, optional.
            If True, check if the connections have duplicate names. Default is True.
    split_ground : bool, optional.
            If True, split the global ground connection to independant ground connections. Default is True.
    split_multi : bool, optional.
            If True, use a splitter to handle connections involving more than two components. This approach
            increases the computational load for individual computations. However, it proves advantageous for
            batch processing by enabling a more comprehensive reduction of circuits, leading to more efficiency
            in batch computations. Default is False.
    max_nports : int, optional.
            The maximum number of ports of a Network that can be reduced in circuit. If a Network in the
            circuit has a number of ports (nports), using the Network.connect() method to reduce the circuit's
            dimensions becomes less efficient compared to directly calculating it with Circuit.s_external.
            This value depends on the performance of the computer and the scale of the circuit. Default is 20.
    dynamic_networks : Sequence[Network], optional.
            A sequence of Networks to ignore in the reduction process. Default is an empty tuple.


    Returns
    -------
    reduced_cnxs : list.
            The reduced connections.


    Examples
    --------
    >>> import skrf as rf
    >>> import numpy as np
    >>> circuit = rf.Circuit(connections)
    >>> reduced_cnxs = rf.reduce_circuit(connections)
    >>> reduced_circuit = rf.Circuit(reduced_cnxs)
    >>> ntwkA = circuit.network
    >>> ntwkB = reduced_circuit.network
    >>> np.allclose(ntwkA.s, ntwkB.s)
    True
    """
    # Pre-processing the dynamic_networks tuple
    ignore_ntwk_names: set[str] = set(ntw.name for ntw in dynamic_networks)

    def invalide_to_reduce(cnx: list[tuple[Network, int]]) -> bool:
        return (
            any(
                (
                    Circuit._is_port(ntwk)
                    or ntwk.nports > max_nports
                    or ntwk.name in ignore_ntwk_names
                )
                for ntwk, _ in cnx
            )
            or len(cnx) != 2
        )

    if split_ground:
        tmp_cnxs = []
        for cnx in connections:
            ground_ntwk = next((ntwk for ntwk, _ in cnx if Circuit._is_ground(ntwk)), None)

            # If there is no ground network or if the connection has exactly 2 elements, append it as is
            if not ground_ntwk or len(cnx) == 2:
                tmp_cnxs.append(cnx)
                continue

            # Otherwise, create new ground connections
            for ntwk, port in cnx:
                if Circuit._is_ground(ntwk):
                    continue
                tmp_gnd = Circuit.Ground(frequency=ground_ntwk.frequency,
                                         name=f'G_{ntwk.name}_{port}')
                tmp_gnd.z0 = ground_ntwk.z0
                tmp_cnxs.append([(ntwk, port), (tmp_gnd, 0)])

        connections = tmp_cnxs

    if split_multi:
        tmp_cnxs = []

        for cnx in connections:
            # Check if the connection has more than 2 components
            if len(cnx) <= 2:
                tmp_cnxs.append(cnx)
                continue

            # Create a splitter for the connection
            _media = media.DefinedGammaZ0(cnx[0][0].frequency)
            splitter = _media.splitter(
                name=f"Splt_{'&'.join(ntwk.name for ntwk, _ in cnx)}",
                nports=len(cnx),
                z0=np.array([ntwk.z0[:, p] for ntwk, p in cnx]).T
            )

            # Connect the splitter to the connection
            for (idx, (ntwk, port)) in enumerate(cnx):
                tmp_cnxs.append([(splitter, idx), (ntwk, port)])

        connections = tmp_cnxs

    # Cache the connections that have been processed to avoid loop
    processed_network_names: set[str] = set()

    # Calculate the total number of Network ports in the specified connection
    def calculate_ports(cnx: list[tuple[Network, int]]) -> int:
        if invalide_to_reduce(cnx):
            return -1

        name_list = sorted(ntwk.name for ntwk, _ in cnx)
        ntwks_str: str = ''.join(name_list)

        unique_networks = len(set(name_list))
        total_ports = sum(ntwk.nports for ntwk, _ in cnx) - 2

        # Return the number of ports if the connections performed
        ports = total_ports if unique_networks == 2 else total_ports // 2 - 1

        # If tuples of Networks in 'connections' have the same Networks, they form a loop.
        # Prioritize processing loops in the circuit by reducing the number of ports by 1.
        # This reduces the computational load during the circuit reduction process.
        if ntwks_str in processed_network_names:
            ports -= 1
        else:
            processed_network_names.add(ntwks_str)

        return ports

    # List of tuples containing connection indices and their calculated ports
    cnx_ports_list = [(idx, calculate_ports(cnx)) for idx, cnx in enumerate(connections)]
    reorder_indices = [idx for idx, _ in sorted(cnx_ports_list, key=lambda x: x[1])]

    # Reorder connections
    connections = [connections[i] for i in reorder_indices]

    # check if the connections are valid
    if check_duplication:
        connections_list = [conn for conn in enumerate(chain.from_iterable(connections))]
        Circuit.check_duplicate_names(connections_list)

    # Use list comprehension to find the connection need to be reduced
    gen = (
        (idx, cnx)
        for idx, cnx in enumerate(connections)
        if not invalide_to_reduce(cnx)
    )

    # Get the first connection need to be reduced
    skip_idx, cnx_to_reduce = next(gen, (-1, [(Network(), -1)] * 2))

    # If there is no connection need to reduce, return the original circuit
    if skip_idx == -1:
        return connections

    # Connect the connections that need to be reduced
    (ntwkA, k), (ntwkB, l) = cnx_to_reduce
    ntwks_name = (ntwkA.name, ntwkB.name)

    # Get the Networks' names
    name_cnt, name_a, name_b = "", str(ntwkA.name), str(ntwkB.name)

    # Generate the connected network and the original port index
    if ntwkA.name == ntwkB.name:
        ntwk_cnt = innerconnect(ntwkA=ntwkA, k=k, l=l)
        name_cnt = f"<{name_a}>"
    else:
        ntwk_cnt = connect(ntwkA=ntwkA, k=k, ntwkB=ntwkB, l=l)
        name_cnt = f"({name_a}**{name_b})"

    # Update the name of the connected network
    ntwk_cnt.name = name_cnt

    # Generate the port index, the index is the original port index
    # and the value is the new port index, -1 means the port is removed.
    port_idx = tuple()
    if ntwkA.name == ntwkB.name:
        port_cnt = list(range(ntwk_cnt.nports))
        port_cnt.insert(min(k, l), -1)
        port_cnt.insert(max(k, l), -1)
        port_idx = (tuple(port_cnt), tuple(port_cnt))
    elif ntwkB.nports == 2 and ntwkA.nports > 2:
        # if ntwkB is a 2port, then keep port indices where you expect.
        port_idx = (
            tuple([(i if i != k else -1) for i in range(ntwkA.nports)]),
            ((-1, k) if l == 0 else (k, -1)),
        )
    else:
        portA = list(range(ntwkA.nports - 1))
        portA.insert(k, -1)
        portB = [i + ntwkA.nports - 1 for i in range(ntwkB.nports - 1)]
        portB.insert(l, -1)

        port_idx = (tuple(portA), tuple(portB))

    # Perform the reduction to get the reduced circuit connections
    # Skip the connection that connected and replace the network and port index
    reduced_cnxs = []
    for idx, cnx in enumerate(connections):
        # Skip the connection reduced
        if idx == skip_idx:
            continue

        tmp_cnx = []
        for ntwk, port in cnx:
            name = ntwk.name
            ntwk_changed = name in ntwks_name

            ntwk_tmp = ntwk
            port_tmp = port

            # Update the connected network and port index
            if ntwk_changed:
                ntwk_tmp = ntwk_cnt
                port_tmp = port_idx[ntwks_name.index(name)][port]

            tmp_cnx.append((ntwk_tmp, port_tmp))

        reduced_cnxs.append(tmp_cnx)

    return reduce_circuit(
        connections=reduced_cnxs,
        check_duplication=False,
        split_ground=False,
        split_multi=False,
        max_nports=max_nports,
        dynamic_networks=dynamic_networks,
    )



"""
.. module:: skrf.vi
========================================================
virtual instruments (:mod:`skrf.vi`)
========================================================

This module defines "virtual instruments" or interfaces to a number of different
types of instruments used in RF measurements and experiments.

Creating A Driver
=================

Drivers are classes inherited from the respective instrument type's abstract base
class and define methods specific to each instrument. This syntax can be complex
and vary wildly between devices and manufacturers. The vi module attempts to
present a standard interface for each.

Basically, creating a driver consists of the following:

1. Creating an instrument class that inherits from its abstract base class
2. Consulting the device's datasheet to find the command for each method in the
   abstract base class

Those wishing to use an instrument that's not supported are encouraged to
consult existing classes to understand the architecture, develop their own
class, and contribute it to scikit-rf.

If you choose to do so, please try to use the following standard terms for the
associated meaning

Conventions
-----------

==================== ========================================================
Name                 Meaning
==================== ========================================================
frequency            Frequency settings [:class:`skrf.Frequency`]
freq_start           Start frequency [Hz]
freq_stop            Stop frequency [Hz]
freq_step            Frequency step [Hz]
freq_center          Center frequency [Hz]
freq_span            Frequency span [Hz]
npoints              Number of frequency points
sweep_time           Time taken for a frequency sweep [s]
sweep_type           How frequency points are distributed (linear,log,etc.)
sweep_mode           How the instrument is triggered (continuous,single,etc)
averaging_on         Whether averaging is on or not
averaging_count      The number of measurements combined for an average
averaging_mode       How averages are taken (per point, each sweep, etc)
if_bandwidth         IF bandwidth [Hz]
==================== ========================================================

Validators
----------

.. autosummary::
    :toctree: generated/

    validators.IntValidator
    validators.FloatValidator
    validators.FreqValidator
    validators.EnumValidator
    validators.SetValidator
    validators.DictValidator
    validators.DelimitedStrValidator
    validators.BooleanValidator

"""
from . import validators



from collections import OrderedDict

colors = OrderedDict()
colors["lime_green"] = "#00FF00"
colors["green"] = "#00AA00"
colors["cyan"] = "#00FFFF"
colors["blue"] = "#0000FF"
colors["red"] = "#FF0000"
colors["magenta"] = "#FF00FF"
colors["yellow"] = "#FFFF00"
colors["purple"] = "#990099"
colors["orange"] = "#FFA500"


def trace_color_cycle(start=0):
    """
    :start n: int
    :return:
    """
    count = start
    color_list = [colors["blue"], colors["red"], colors["magenta"], colors["green"]]
    num = len(color_list)
    while count < 1000:
        yield color_list[count % num]
        count += 1



"""
.. module:: skrf.vi.validators
==============================
validators (:mod:`skrf.vi.validators`)
==============================

.. autosummary::
"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Sequence

import re
from abc import ABC, abstractmethod
from enum import Enum


class ValidationError(Exception):
    pass


class Validator(ABC):
    @abstractmethod
    def validate_input(self, arg) -> Any:
        """Try to convert arg to a valid value and return. Used to ensure
        commands sent to the instrument are the proper format."""
        pass

    def validate_output(self, arg) -> Any:
        """Try to convert arg to a valid value and return. Used to convert
        responses from the instrument to something useful. By default,
        this just calls validate_input, but can be overwritten in subclasses.
        (See EnumValidator)"""
        return self.validate_input(arg)


class IntValidator(Validator):
    def __init__(self, min: int | None = None, max: int | None = None) -> None:
        self.min = min
        self.max = max

    def validate_input(self, arg) -> int:
        try:
            arg = int(arg)
        except ValueError as err:
            raise ValidationError(f"Could not convert {arg} to an int") from err

        if self.min is not None or self.max is not None:
            self.check_bounds(arg)

        return arg

    def check_bounds(self, arg):
        try:
            if self.min is not None:
                assert arg >= self.min
        except AssertionError as err:
            raise ValidationError(f"{arg} < {self.min}") from err

        try:
            if self.max is not None:
                assert arg <= self.max
        except AssertionError as err:
            raise ValidationError(f"{arg} > {self.max}") from err


class FloatValidator(Validator):
    def __init__(
        self,
        min: float | None = None,
        max: float | None = None,
        decimal_places: int=50
    ) -> None:
        self.min = min
        self.max = max
        self.dec_places = decimal_places

    def validate_input(self, arg) -> float:
        try:
            arg = float(arg)
        except ValueError as err:
            raise ValidationError(f"Could not convert {arg} to a float") from err

        if self.min is not None or self.max is not None:
            self.check_bounds(arg)

        return round(arg, self.dec_places)

    def check_bounds(self, arg):
        try:
            if self.min is not None:
                assert arg >= self.min
        except AssertionError as err:
            raise ValidationError(f"{arg} < {self.min}") from err

        try:
            if self.max is not None:
                assert arg <= self.max
        except AssertionError as err:
            raise ValidationError(f"{arg} > {self.max}") from err


class FreqValidator(Validator):
    freq_re = re.compile(r"(?P<val>\d+\.?\d*)\s*(?P<si_prefix>[kMG])?(?:[hH][zZ])?")
    si = {"k": 1e3, "M": 1e6, "G": 1e9}

    def validate_input(self, arg) -> int:
        if isinstance(arg, str):
            f = re.fullmatch(self.freq_re, arg)
            if f is None:
                raise ValidationError("Invalid frequency string")
            return int(float(f["val"]) * self.si.get(f["si_prefix"], 1))

        else:
            try:
                return int(arg)
            except ValueError as err:
                raise ValidationError("Could not convert {arg} to an int") from err

    def validate_output(self, arg) -> int:
        try:
            f = float(arg)
            return int(f)
        except ValueError as err:
            raise ValidationError(f"Response from instrument ({arg}) could not be converted to an int") from err


class EnumValidator(Validator):
    def __init__(self, enum: Enum) -> None:
        self.Enum = enum

    def validate_input(self, arg) -> Any:
        if isinstance(arg, Enum):
            return arg.value
        else:
            try:
                return self.Enum(arg).value
            except ValueError as err:
                raise ValidationError(f"{arg} is not a valid {self.Enum.__name__}") from err

    def validate_output(self, arg) -> Any:
        try:
            return self.Enum(arg)
        except ValueError as err:
            raise ValidationError(f"Got unexpected response {arg}") from err


class SetValidator(Validator):
    def __init__(self, valid: Sequence) -> None:
        dtype = type(valid[0])
        if not all(isinstance(x, dtype) for x in valid):
            raise ValueError("All elements of set must be of the same type.")

        self.valid = valid
        self.dtype = dtype

    def validate_input(self, arg) -> Any:
        arg = self.dtype(arg)
        if arg in self.valid:
            return arg
        else:
            raise ValidationError(f"{arg} is not in {self.valid}")


class DictValidator(Validator):
    def __init__(self, arg_string: str, response_pattern: re.Pattern | str) -> None:
        self.arg_string = arg_string
        if isinstance(response_pattern, str):
            self.pattern = re.compile(response_pattern)
        else:
            self.pattern = response_pattern

    def validate_input(self, arg) -> str:
        try:
            return self.arg_string.format(**arg)
        except KeyError as e:
            raise ValidationError(f"Missing expected argument '{e}'") from e

    def validate_output(self, arg) -> dict:
        match = self.pattern.fullmatch(arg)
        if match:
            return match.groupdict()
        else:
            raise ValidationError(
                "Response did not fit regex. "
                f"Response: {arg} Pattern: {self.pattern.pattern}"
            )

class DelimitedStrValidator(Validator):
    def __init__(self, dtype: type =str , sep: str = ',') -> None:
        self.dtype = dtype
        self.sep = sep

    def validate_input(self, arg: Sequence) -> str:
        if not all(isinstance(x, self.dtype) for x in arg):
            raise ValidationError("All elements must be of type {self.dtype}.")

        return self.sep.join(str(x) for x in arg)

    def validate_output(self, arg: str) -> list:
        arg = arg.replace('"', '')
        return [self.dtype(val) for val in arg.split(self.sep)]

class BooleanValidator(Validator):
    truthy = ['1', 'on', 'true']
    falsey = ['0', 'off', 'false']

    def __init__(
        self,
        true_response: str | None = None,
        false_response: str | None = None,
        true_setting: str='1',
        false_setting: str='0'
    ):
        if true_response:
            self.truthy.append(true_response.lower())
        if false_response:
            self.falsey.append(false_response.lower())

        self.true_val = true_setting
        self.false_val = false_setting

    def validate_input(self, arg) -> str:
        if str(arg).lower() in self.truthy:
            return self.true_val
        elif str(arg).lower() in self.falsey:
            return self.false_val
        else:
            raise ValidationError('Argument must be a truthy or falsey value')

    def validate_output(self, arg) -> bool:
        return str(arg).lower() in self.truthy
